# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any, List

from pyhocon import ConfigFactory, ConfigTree

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_utils import ModeDashboardUtils
from databuilder.rest_api.mode_analytics.mode_paginated_rest_api_query import ModePaginatedRestApiQuery
from databuilder.rest_api.query_merger import QueryMerger
from databuilder.transformer.base_transformer import ChainedTransformer, Transformer
from databuilder.transformer.dict_to_model import MODEL_CLASS, DictToModel
from databuilder.transformer.template_variable_substitution_transformer import (
    FIELD_NAME as VAR_FIELD_NAME, TEMPLATE, TemplateVariableSubstitutionTransformer,
)
from databuilder.transformer.timestamp_string_to_epoch import FIELD_NAME, TimestampStringToEpoch

LOGGER = logging.getLogger(__name__)

# a list of space tokens that we want to skip indexing
DASHBOARD_GROUP_IDS_TO_SKIP = 'dashboard_group_ids_to_skip'


class ModeDashboardExtractor(Extractor):
    """
    A Extractor that extracts core metadata on Mode dashboard. https://app.mode.com/
    It extracts list of reports that consists of:
        Dashboard group name (Space name)
        Dashboard group id (Space token)
        Dashboard group description (Space description)
        Dashboard name (Report name)
        Dashboard id (Report token)
        Dashboard description (Report description)

    Other information such as report run, owner, chart name, query name is in separate extractor.
    """

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf

        self.dashboard_group_ids_to_skip = self._conf.get_list(DASHBOARD_GROUP_IDS_TO_SKIP, [])

        restapi_query = self._build_restapi_query()
        self._extractor = ModeDashboardUtils.create_mode_rest_api_extractor(restapi_query=restapi_query,
                                                                            conf=self._conf)

        # Payload from RestApiQuery has timestamp which is ISO8601. Here we are using TimestampStringToEpoch to
        # transform into epoch and then using DictToModel to convert Dictionary to Model
        transformers: List[Transformer] = []
        timestamp_str_to_epoch_transformer = TimestampStringToEpoch()
        timestamp_str_to_epoch_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, timestamp_str_to_epoch_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict({FIELD_NAME: 'created_timestamp', })))

        transformers.append(timestamp_str_to_epoch_transformer)

        dashboard_group_url_transformer = TemplateVariableSubstitutionTransformer()
        dashboard_group_url_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, dashboard_group_url_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict({VAR_FIELD_NAME: 'dashboard_group_url',
                                         TEMPLATE: 'https://app.mode.com/{organization}/spaces/{dashboard_group_id}'})))

        transformers.append(dashboard_group_url_transformer)

        dashboard_url_transformer = TemplateVariableSubstitutionTransformer()
        dashboard_url_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, dashboard_url_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict({VAR_FIELD_NAME: 'dashboard_url',
                                         TEMPLATE: 'https://app.mode.com/{organization}/reports/{dashboard_id}'})))
        transformers.append(dashboard_url_transformer)

        dict_to_model_transformer = DictToModel()
        dict_to_model_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, dict_to_model_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict(
                    {MODEL_CLASS: 'databuilder.models.dashboard.dashboard_metadata.DashboardMetadata'})))
        transformers.append(dict_to_model_transformer)

        self._transformer = ChainedTransformer(transformers=transformers)

    def extract(self) -> Any:
        record = self._extractor.extract()

        # determine whether we want to skip these records
        while record and record.get('dashboard_group_id') in self.dashboard_group_ids_to_skip:
            record = self._extractor.extract()

        if not record:
            return None

        return self._transformer.transform(record=record)

    def get_scope(self) -> str:
        return 'extractor.mode_dashboard'

    def _build_restapi_query(self) -> ModePaginatedRestApiQuery:
        """
        Build REST API Query to get Mode Dashboard metadata
        :return: A RestApiQuery that provides Mode Dashboard metadata
        """
        seed_query = ModeDashboardUtils.get_seed_query(conf=self._conf)
        params = ModeDashboardUtils.get_auth_params(conf=self._conf, discover_auth=True)

        # Reports
        # https://mode.com/developer/discovery-api/analytics/reports/
        url = 'https://app.mode.com/batch/{organization}/reports'
        json_path = 'reports[*].[token, name, description, created_at, space_token]'
        field_names = ['dashboard_id', 'dashboard_name', 'description', 'created_timestamp', 'dashboard_group_id']
        max_record_size = 1000
        pagination_json_path = 'reports[*]'

        spaces_query = ModeDashboardUtils.get_spaces_query_api(conf=self._conf)
        query_merger = QueryMerger(query_to_merge=spaces_query, merge_key='dashboard_group_id')

        reports_query = ModePaginatedRestApiQuery(query_to_join=seed_query, url=url, params=params,
                                                  json_path=json_path, field_names=field_names,
                                                  skip_no_result=True, max_record_size=max_record_size,
                                                  pagination_json_path=pagination_json_path,
                                                  query_merger=query_merger)

        return reports_query



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging

from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_executions_extractor import (
    ModeDashboardExecutionsExtractor,
)
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_utils import ModeDashboardUtils
from databuilder.extractor.restapi.rest_api_extractor import STATIC_RECORD_DICT
from databuilder.rest_api.mode_analytics.mode_paginated_rest_api_query import ModePaginatedRestApiQuery
from databuilder.transformer.dict_to_model import MODEL_CLASS, DictToModel
from databuilder.transformer.timestamp_string_to_epoch import FIELD_NAME, TimestampStringToEpoch

LOGGER = logging.getLogger(__name__)


class ModeDashboardLastModifiedTimestampExtractor(ModeDashboardExecutionsExtractor):
    """
    A Extractor that extracts Mode dashboard's last modified timestamp.

    """

    def __init__(self) -> None:
        super(ModeDashboardLastModifiedTimestampExtractor, self).__init__()

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(
            ConfigFactory.from_dict({
                STATIC_RECORD_DICT: {'product': 'mode'},
                f'{DictToModel().get_scope()}.{MODEL_CLASS}':
                    'databuilder.models.dashboard.dashboard_last_modified.DashboardLastModifiedTimestamp',
                f'{TimestampStringToEpoch().get_scope()}.{FIELD_NAME}':
                    'last_modified_timestamp'
            })
        )
        super(ModeDashboardLastModifiedTimestampExtractor, self).init(conf)

    def get_scope(self) -> str:
        return 'extractor.mode_dashboard_last_modified_timestamp_execution'

    def _build_restapi_query(self) -> ModePaginatedRestApiQuery:
        """
        Build REST API Query to get Mode Dashboard last modified timestamp
        :return: A RestApiQuery that provides Mode Dashboard last successful execution (run)
        """

        seed_query = ModeDashboardUtils.get_seed_query(conf=self._conf)
        params = ModeDashboardUtils.get_auth_params(conf=self._conf, discover_auth=True)

        # Reports
        # https://mode.com/developer/discovery-api/analytics/reports/
        url = 'https://app.mode.com/batch/{organization}/reports'
        json_path = 'reports[*].[token, space_token, edited_at]'
        field_names = ['dashboard_id', 'dashboard_group_id', 'last_modified_timestamp']
        max_record_size = 1000
        pagination_json_path = 'reports[*]'
        last_modified_query = ModePaginatedRestApiQuery(query_to_join=seed_query, url=url, params=params,
                                                        json_path=json_path, field_names=field_names,
                                                        skip_no_result=True, max_record_size=max_record_size,
                                                        pagination_json_path=pagination_json_path)

        return last_modified_query



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import hashlib
from typing import (
    Iterator, List, Optional,
)

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.query.base import QueryBase
from databuilder.models.query.query import QueryMetadata
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata


class QueryWhereMetadata(QueryBase):
    """
    A Where clause used on a query.
    """
    NODE_LABEL = 'Where'
    KEY_FORMAT = '{table_hash}-{where_hash}'

    # Relation between table and query
    COLUMN_WHERE_RELATION_TYPE = 'USES_WHERE_CLAUSE'
    INVERSE_COLUMN_WHERE_RELATION_TYPE = 'WHERE_CLAUSE_USED_ON'

    QUERY_WHERE_RELATION_TYPE = 'HAS_WHERE_CLAUSE'
    INVERSE_QUERY_WHERE_RELATION_TYPE = 'WHERE_CLAUSE_OF'

    # Node attributes
    WHERE_CLAUSE = 'where_clause'
    LEFT_ARG = 'left_arg'
    RIGHT_ARG = 'right_arg'
    OPERATOR = 'operator'
    ALIAS_MAPPING = 'alias_mapping'

    def __init__(self,
                 tables: List[TableMetadata],
                 where_clause: str,
                 left_arg: Optional[str],
                 right_arg: Optional[str],
                 operator: Optional[str],
                 query_metadata: Optional[QueryMetadata] = None,
                 yield_relation_nodes: bool = False):
        """
        :param tables: List of table meteadata objects corresponding to tables in this where clause
        :param where_clause: a sting representation of the SQL where clause
        :param left_arg: An optional string representing the left side of the where cause, e.g.
            in the clause (where x < 3), this would be "x"
        :param operator: An optional string representing the operator in the where cause, e.g.
            in the clause (where x < 3), this would be "<"
        :param right_arg: An optional string representing the right side of the where cause, e.g.
            in the clause (where x < 3), this would be "3"
        :param query_metadata: The Query metadata object that this where clause belongs to, this
            is optional to allow creating static QueryWhereMetadata objects to show on tables
            without the complexity of creating QueryMetadata
        :param yield_relation_nodes: A boolean, indicating whether or not the query metadata
            and tables associated to this Where should have nodes created if they does not
            already exist.
        """
        self.tables = tables
        self.query_metadata = query_metadata
        self.where_clause = where_clause
        self.left_arg = left_arg
        self.right_arg = right_arg
        self.operator = operator
        self.yield_relation_nodes = yield_relation_nodes
        self._table_hash = self._get_table_hash(self.tables)
        self._where_hash = self._get_where_hash(self.where_clause)
        self._node_iter = self._create_next_node()
        self._relation_iter = self._create_relation_iterator()

    def __repr__(self) -> str:
        tbl_str = self.tables[0]._get_table_key()
        if len(self.tables) > 1:
            tbl_str += f' + {len(self.tables) - 1} other tables'
        return f'QueryWhereMetadata(Table: {tbl_str}, {self.where_clause[:25]})'

    def _get_table_hash(self, tables: List[TableMetadata]) -> str:
        """
        Generates a unique hash for a set of tables that are associated to a where clause. Since
        we do not want multiple instances of this where clause represented in the database we may
        need to link mulitple tables to this where clause. We do this by creating a single, unique
        key across multiple tables by concatenating all of the table keys together and creating a
        hash (to shorten the value).
        """
        tbl_keys = ''.join(list(sorted([t._get_table_key() for t in tables])))
        return hashlib.md5(tbl_keys.encode('utf-8')).hexdigest()

    def _get_where_hash(self, where_clause: str) -> str:
        """
        Generates a unique hash for a where clause.
        """
        sql_no_fmt = self._normalize(where_clause)
        return hashlib.md5(sql_no_fmt.encode('utf-8')).hexdigest()

    def create_next_node(self) -> Optional[GraphNode]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    @staticmethod
    def get_key(table_hash: str, where_hash: str) -> str:
        return QueryWhereMetadata.KEY_FORMAT.format(table_hash=table_hash, where_hash=where_hash)

    def get_key_self(self) -> str:
        return QueryWhereMetadata.get_key(table_hash=self._table_hash, where_hash=self._where_hash)

    def get_query_relations(self) -> Iterator[GraphRelationship]:
        for table in self.tables:
            for col in table.columns:
                yield GraphRelationship(
                    start_label=ColumnMetadata.COLUMN_NODE_LABEL,
                    end_label=self.NODE_LABEL,
                    start_key=table._get_col_key(col),
                    end_key=self.get_key_self(),
                    type=self.COLUMN_WHERE_RELATION_TYPE,
                    reverse_type=self.INVERSE_COLUMN_WHERE_RELATION_TYPE,
                    attributes={}
                )

        # Optional Query to Where Clause
        if self.query_metadata:
            yield GraphRelationship(
                start_label=QueryMetadata.NODE_LABEL,
                end_label=self.NODE_LABEL,
                start_key=self.query_metadata.get_key_self(),
                end_key=self.get_key_self(),
                type=self.QUERY_WHERE_RELATION_TYPE,
                reverse_type=self.INVERSE_QUERY_WHERE_RELATION_TYPE,
                attributes={}
            )

    def _create_next_node(self) -> Iterator[GraphNode]:
        """
        Create query nodes
        :return:
        """
        yield GraphNode(
            key=self.get_key_self(),
            label=self.NODE_LABEL,
            attributes={
                self.WHERE_CLAUSE: self.where_clause,
                self.LEFT_ARG: self.left_arg,
                self.RIGHT_ARG: self.right_arg,
                self.OPERATOR: self.operator
            }
        )
        if self.yield_relation_nodes:
            for table in self.tables:
                for tbl_item in table._create_next_node():
                    yield tbl_item
            if self.query_metadata:
                for query_item in self.query_metadata._create_next_node():
                    yield query_item

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        relations = self.get_query_relations()
        for relation in relations:
            yield relation

        if self.yield_relation_nodes:
            for table in self.tables:
                for tbl_rel in table._create_next_relation():
                    yield tbl_rel
            if self.query_metadata:
                for query_rel in self.query_metadata._create_relation_iterator():
                    yield query_rel



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
import json
from typing import (
    Dict, Iterator, Optional, Union,
)

from databuilder.extractor.es_base_extractor import ElasticsearchBaseExtractor
from databuilder.models.table_metadata import TableMetadata


class ElasticsearchMetadataExtractor(ElasticsearchBaseExtractor):
    """
    Extractor to extract index metadata from Elasticsearch
    """

    def get_scope(self) -> str:
        return 'extractor.es_metadata'

    def _render_programmatic_description(self, input: Optional[Dict]) -> Optional[str]:
        if input:
            result = f"""```\n{json.dumps(input, indent=2)}\n```"""

            return result
        else:
            return None

    def _get_extract_iter(self) -> Iterator[Union[TableMetadata, None]]:
        indexes: Dict = self._get_indexes()

        for index_name, index_metadata in indexes.items():
            properties = self._get_index_mapping_properties(index_metadata) or dict()

            columns = self._get_attributes(input_mapping=properties)

            # The columns are already sorted, but the sort_order needs to be added to each column metadata entry
            if self._correct_sort_order:
                for index in range(len(columns)):
                    columns[index].sort_order = index

            table_metadata = TableMetadata(database=self.database,
                                           cluster=self.cluster,
                                           schema=self.schema,
                                           name=index_name,
                                           description=None,
                                           columns=columns,
                                           is_view=False,
                                           tags=None,
                                           description_source=None)

            yield table_metadata

            if self._extract_technical_details:
                _settings = index_metadata.get('settings', dict())
                _aliases = index_metadata.get('aliases', dict())

                settings = self._render_programmatic_description(_settings)
                aliases = self._render_programmatic_description(_aliases)

                if aliases:
                    yield TableMetadata(database=self.database,
                                        cluster=self.cluster,
                                        schema=self.schema,
                                        name=index_name,
                                        description=aliases,
                                        columns=columns,
                                        is_view=False,
                                        tags=None,
                                        description_source='aliases')
                if settings:
                    yield TableMetadata(database=self.database,
                                        cluster=self.cluster,
                                        schema=self.schema,
                                        name=index_name,
                                        description=settings,
                                        columns=columns,
                                        is_view=False,
                                        tags=None,
                                        description_source='settings')



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Union  # noqa: F401

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship

NODE_KEY = 'KEY'
NODE_LABEL = 'LABEL'

RELATION_START_KEY = 'START_KEY'
RELATION_START_LABEL = 'START_LABEL'
RELATION_END_KEY = 'END_KEY'
RELATION_END_LABEL = 'END_LABEL'
RELATION_TYPE = 'TYPE'
RELATION_REVERSE_TYPE = 'REVERSE_TYPE'


class GraphSerializable(object, metaclass=abc.ABCMeta):
    """
    A Serializable abstract class asks subclass to implement next node or
    next relation in dict form so that it can be serialized to CSV file.

    Any model class that needs to be pushed to a graph database should inherit this class.
    """

    def __init__(self) -> None:
        pass

    @abc.abstractmethod
    def create_next_node(self) -> Union[GraphNode, None]:
        """
        Creates GraphNode the process that consumes this class takes the output
        serializes to the desired graph database.

        :return: a GraphNode or None if no more records to serialize
        """
        raise NotImplementedError

    @abc.abstractmethod
    def create_next_relation(self) -> Union[GraphRelationship, None]:
        """
        Creates GraphRelationship the process that consumes this class takes the output
        serializes to the desired graph database.

        :return: a GraphRelationship or None if no more record to serialize
        """
        raise NotImplementedError

    def next_node(self) -> Union[GraphNode, None]:
        node_dict = self.create_next_node()
        if not node_dict:
            return None

        self._validate_node(node_dict)
        return node_dict

    def next_relation(self) -> Union[GraphRelationship, None]:
        relation_dict = self.create_next_relation()
        if not relation_dict:
            return None

        self._validate_relation(relation_dict)
        return relation_dict

    def _validate_node(self, node: GraphNode) -> None:
        node_id, node_label, _ = node

        if node_id is None:
            raise RuntimeError('Required header missing. Required attributes id and label , Missing: id')

        if node_label is None:
            raise RuntimeError('Required header missing. Required attributes id and label , Missing: label')

        self._validate_label_value(node_label)

    def _validate_relation(self, relation: GraphRelationship) -> None:
        self._validate_label_value(relation.start_label)
        self._validate_label_value(relation.end_label)
        self._validate_relation_type_value(relation.type)
        self._validate_relation_type_value(relation.reverse_type)

    def _validate_relation_type_value(self, value: str) -> None:
        if not value.isupper():
            raise RuntimeError(f'TYPE needs to be upper case: {value}')

    def _validate_label_value(self, value: str) -> None:
        if not value.istitle():
            raise RuntimeError(f'LABEL should only have upper case character on its first one: {value}')



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import os
from enum import Enum
from typing import (
    Dict, Iterator, List, Optional, Tuple, Union,
)

from pyhocon import ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.badge import Badge, BadgeMetadata
from databuilder.models.table_lineage import TableLineage
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata
from databuilder.models.table_source import TableSource

LOGGER = logging.getLogger(__name__)


DBT_CATALOG_REQD_KEYS = ['nodes']
DBT_MANIFEST_REQD_KEYS = ['nodes', 'child_map']
DBT_MODEL_TYPE = 'model'
DBT_MODEL_PREFIX = 'model.'
DBT_TEST_PREFIX = 'test.'


class DBT_TAG_AS(Enum):
    BADGE = 'badge'
    TAG = 'tag'


class DBT_MODEL_NAME_KEY(Enum):
    ALIAS = 'alias'
    NAME = 'name'


class InvalidDbtInputs(Exception):
    pass


class DbtExtractor(Extractor):
    """
    Extracts metadata from the dbt manifest.json and catalog.json files.
    At least one of a manifest or a catalog (or both) must be provided.
    The location of the file or a valid Python dictionary of the content
    can be provided.

    Currently the following assets are extracted from these files:

    - Tables
    - Columns
    - Definitions
    - Table lineage
    - Tags (converted to Amundsen Badges)

    Additional metadagta exists and may be extracted in the future:

    - Run / test outcomes
    - Freshness
    - Hooks (as programatic description?)
    - Analysis (as queries for a table??)
    - Table / column level statistics
    - Table comments (as programatic descriptoins)
    """

    CATALOG_JSON = "catalog_json"
    MANIFEST_JSON = "manifest_json"
    DATABASE_NAME = 'database_name'

    # Dbt Extract Options
    EXTRACT_TABLES = 'extract_tables'
    EXTRACT_DESCRIPTIONS = 'extract_descriptions'
    EXTRACT_TAGS = 'extract_tags'
    EXTRACT_LINEAGE = 'extract_lineage'
    SOURCE_URL = 'source_url'  # Base source code URL for the repo containing dbt workflows
    IMPORT_TAGS_AS = 'import_tags_as'
    SCHEMA_FILTER = 'schema_filter'  # Only extract dbt models from this schema, defaults to all models
    MODEL_NAME_KEY = 'model_name_key'  # Whether to use the "name" or "alias" from dbt as the Amundsen name

    # Makes all db, schema, cluster and table names lowercase. This is done so that table metadata from dbt
    # with the default key `Sample://Cluster/Schema/Table` match existing metadata that Amundsen has from
    # the database, which may be `sample://cluster/schema/table`.
    # Most databases that dbt integrates with either use lowercase by default in the information schema
    # or the default Amundsen extractor applies a `lower(...)` function to the result (e.g. snowflake).
    # However, Amundsen does not currently enforce a consistent convention and some databases do support
    # upper and lowercase naming conventions (e.g. Redshift). It may be useful to set this False in the
    # config if the table metadata keys in your database are not all lowercase and to then use a transformer to
    # properly format the string value.
    FORCE_TABLE_KEY_LOWER = 'force_table_key_lower'

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf
        self._database_name = conf.get_string(DbtExtractor.DATABASE_NAME)
        self._dbt_manifest = conf.get_string(DbtExtractor.MANIFEST_JSON)
        self._dbt_catalog = conf.get_string(DbtExtractor.CATALOG_JSON)
        # Extract options
        self._extract_tables = conf.get_bool(DbtExtractor.EXTRACT_TABLES, True)
        self._extract_descriptions = conf.get_bool(DbtExtractor.EXTRACT_DESCRIPTIONS, True)
        self._extract_tags = conf.get_bool(DbtExtractor.EXTRACT_TAGS, True)
        self._extract_lineage = conf.get_bool(DbtExtractor.EXTRACT_LINEAGE, True)
        self._source_url = conf.get_string(DbtExtractor.SOURCE_URL, None)
        self._force_table_key_lower = conf.get_bool(DbtExtractor.FORCE_TABLE_KEY_LOWER, True)
        self._dbt_tag_as = DBT_TAG_AS(conf.get_string(DbtExtractor.IMPORT_TAGS_AS, DBT_TAG_AS.BADGE.value))
        self._schema_filter = conf.get_string(DbtExtractor.SCHEMA_FILTER, '')
        self._model_name_key = DBT_MODEL_NAME_KEY(
            conf.get_string(DbtExtractor.MODEL_NAME_KEY, DBT_MODEL_NAME_KEY.NAME.value)).value
        self._clean_inputs()
        self._extract_iter: Union[None, Iterator] = None

    def get_scope(self) -> str:
        return "extractor.dbt"

    def _validate_catalog(self) -> None:
        # Load the catalog file if needed and run basic validation on the content
        try:
            self._dbt_catalog = json.loads(self._dbt_catalog)
        except Exception:
            try:
                with open(self._dbt_catalog, 'rb') as f:
                    self._dbt_catalog = json.loads(f.read().lower())
            except Exception as e:
                raise InvalidDbtInputs(
                    'Invalid content for a dbt catalog was provided. Must be a valid Python '
                    'dictionary or the location of a file. Error received: %s' % e
                )
        for catalog_key in DBT_CATALOG_REQD_KEYS:
            if catalog_key not in self._dbt_catalog:
                raise InvalidDbtInputs(
                    "Dbt catalog file must contain keys: %s, found keys: %s"
                    % (DBT_CATALOG_REQD_KEYS, self._dbt_catalog.keys())
                )

    def _validate_manifest(self) -> None:
        # Load the manifest file if needed and run basic validation on the content
        try:
            self._dbt_manifest = json.loads(self._dbt_manifest)
        except Exception:
            try:
                with open(self._dbt_manifest, 'rb') as f:
                    self._dbt_manifest = json.loads(f.read().lower())
            except Exception as e:
                raise InvalidDbtInputs(
                    'Invalid content for a dbt manifest was provided. Must be a valid Python '
                    'dictionary or the location of a file. Error received: %s' % e
                )
        for manifest_key in DBT_MANIFEST_REQD_KEYS:
            if manifest_key not in self._dbt_manifest:
                raise InvalidDbtInputs(
                    "Dbt manifest file must contain keys: %s, found keys: %s"
                    % (DBT_MANIFEST_REQD_KEYS, self._dbt_manifest.keys())
                )

    def _clean_inputs(self) -> None:
        """
        Validates the dbt input to ensure that at least one of the inputs
        (manifest.json or catalog.json) are provided. Once validated, the
        inputs are sanitized to ensure that the `self._dbt_manifest` and
        `self._dbt_catalog` are valid Python dictionaries.
        """
        if self._database_name is None:
            raise InvalidDbtInputs(
                'Must provide a database name that corresponds to this dbt catalog and manifest.'
            )

        if not self._dbt_manifest or not self._dbt_catalog:
            raise InvalidDbtInputs(
                'Must provide a dbt manifest file and dbt catalog file.'
            )

        self._validate_catalog()
        self._validate_manifest()

    def extract(self) -> Union[TableMetadata, None]:
        """
        For every feature table from Feast, a multiple objets are extracted:

        1. TableMetadata with feature table description
        2. Programmatic Description of the feature table, containing
           metadata - date of creation and labels
        3. Programmatic Description with Batch Source specification
        4. (if applicable) Programmatic Description with Stream Source
           specification
        """
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def _default_sanitize(self, s: str) -> str:
        """
        Default function that will be run to convert the value of a string to lowercase.
        """
        if s and self._force_table_key_lower:
            s = s.lower()
        return s

    def _get_table_descriptions(self, manifest_content: Dict) -> Tuple[Optional[str], Optional[str]]:
        """
        Gets a description and description source for a table.
        """
        desc, desc_src = None, None
        if self._extract_descriptions:
            desc = manifest_content.get('description')
            desc_src = 'dbt description'
        return desc, desc_src

    def _get_table_tags_badges(self, manifest_content: Dict) -> Tuple[Optional[List[str]], Optional[List[str]]]:
        """
        Gets tags or badges for a given table. At most one of these values will not be null.
        """
        tags, tbl_badges = None, None
        if self._extract_tags:
            if self._dbt_tag_as == DBT_TAG_AS.BADGE:
                tbl_badges = manifest_content.get('tags')
            elif self._dbt_tag_as == DBT_TAG_AS.TAG:
                tags = manifest_content.get('tags')
        return tags, tbl_badges

    def _can_yield_schema(self, schema: str) -> bool:
        """
        Whether or not the schema can be yielded based on the schema filter criteria.
        """
        return (not self._schema_filter) or (self._schema_filter.lower() == schema.lower())

    def _get_extract_iter(self) -> Iterator[Union[TableMetadata, BadgeMetadata, TableSource, TableLineage]]:
        """
        Generates the extract iterator for all of the model types created by the dbt files.
        """
        dbt_id_to_table_key = {}
        for tbl_node, manifest_content in self._dbt_manifest['nodes'].items():

            if manifest_content['resource_type'] == DBT_MODEL_TYPE and tbl_node in self._dbt_catalog['nodes']:
                LOGGER.info(
                    'Extracting dbt {}.{}'.format(manifest_content['schema'], manifest_content[self._model_name_key])
                )

                catalog_content = self._dbt_catalog['nodes'][tbl_node]

                tbl_columns: List[ColumnMetadata] = self._get_column_values(
                    manifest_columns=manifest_content['columns'], catalog_columns=catalog_content['columns']
                )

                desc, desc_src = self._get_table_descriptions(manifest_content)
                tags, tbl_badges = self._get_table_tags_badges(manifest_content)

                tbl_metadata = TableMetadata(
                    database=self._default_sanitize(self._database_name),
                    # The dbt "database" is the cluster here
                    cluster=self._default_sanitize(manifest_content['database']),
                    schema=self._default_sanitize(manifest_content['schema']),
                    name=self._default_sanitize(manifest_content[self._model_name_key]),
                    is_view=catalog_content['metadata']['type'] == 'view',
                    columns=tbl_columns,
                    tags=tags,
                    description=desc,
                    description_source=desc_src
                )
                # Keep track for Lineage
                dbt_id_to_table_key[tbl_node] = tbl_metadata._get_table_key()

                # Optionally filter schemas in the output
                yield_schema = self._can_yield_schema(manifest_content['schema'])

                if self._extract_tables and yield_schema:
                    yield tbl_metadata

                if self._extract_tags and tbl_badges and yield_schema:
                    yield BadgeMetadata(start_label=TableMetadata.TABLE_NODE_LABEL,
                                        start_key=tbl_metadata._get_table_key(),
                                        badges=[Badge(badge, 'table') for badge in tbl_badges])

                if self._source_url and yield_schema:
                    yield TableSource(db_name=tbl_metadata.database,
                                      cluster=tbl_metadata.cluster,
                                      schema=tbl_metadata.schema,
                                      table_name=tbl_metadata.name,
                                      source=os.path.join(self._source_url, manifest_content.get('original_file_path')))

        if self._extract_lineage:
            for upstream, downstreams in self._dbt_manifest['child_map'].items():
                if upstream not in dbt_id_to_table_key:
                    continue
                valid_downstreams = [
                    dbt_id_to_table_key[k] for k in downstreams
                    if k.startswith(DBT_MODEL_PREFIX) and dbt_id_to_table_key.get(k)
                ]
                if valid_downstreams:
                    yield TableLineage(
                        table_key=dbt_id_to_table_key[upstream],
                        downstream_deps=valid_downstreams
                    )

    def _get_column_values(self, manifest_columns: Dict, catalog_columns: Dict) -> List[ColumnMetadata]:
        """
        Iterates over the columns in the manifest file and creates a `ColumnMetadata` object
        with the combined information from the manifest file as well as the catalog file.

        :params manifest_columns: A dictionary of values from the manifest.json, the keys
            are column names and the values are column metadata
        :params catalog_columns: A dictionary of values from the catalog.json, the keys
            are column names and the values are column metadata
        :returns: A list of `ColumnMetadata` in Amundsen.
        """
        tbl_columns = []
        for catalog_col_name, catalog_col_content in catalog_columns.items():
            manifest_col_content = manifest_columns.get(catalog_col_name, {})
            if catalog_col_content:
                col_desc = None
                if self._extract_descriptions:
                    col_desc = manifest_col_content.get('description')

                # Only extract column-level tags IF converting to badges, Amundsen does not have column-level tags
                badges = None
                if self._extract_tags and self._dbt_tag_as == DBT_TAG_AS.BADGE:
                    badges = manifest_col_content.get('tags')

                col_metadata = ColumnMetadata(
                    name=self._default_sanitize(catalog_col_content['name']),
                    description=col_desc,
                    col_type=catalog_col_content['type'],
                    sort_order=catalog_col_content['index'],
                    badges=badges
                )
                tbl_columns.append(col_metadata)
        return tbl_columns



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (  # noqa: F401
    Any, Dict, Iterator, Union,
)

from pyhocon import ConfigFactory, ConfigTree  # noqa: F401

from databuilder.extractor.base_postgres_metadata_extractor import BasePostgresMetadataExtractor


class PostgresMetadataExtractor(BasePostgresMetadataExtractor):
    """
    Extracts Postgres table and column metadata from underlying meta store database using SQLAlchemyExtractor
    """

    def get_sql_statement(self, use_catalog_as_cluster_name: bool, where_clause_suffix: str) -> str:
        if use_catalog_as_cluster_name:
            cluster_source = "current_database()"
        else:
            cluster_source = f"'{self._cluster}'"

        return """
            SELECT
                {cluster_source} as cluster,
                st.schemaname as schema,
                st.relname as name,
                pgtd.description as description,
                att.attname as col_name,
                pgtyp.typname as col_type,
                pgcd.description as col_description,
                att.attnum as col_sort_order
            FROM pg_catalog.pg_attribute att
            INNER JOIN
                pg_catalog.pg_statio_all_tables as st
                on att.attrelid=st.relid
            LEFT JOIN
                pg_catalog.pg_type pgtyp
                on pgtyp.oid=att.atttypid
            LEFT JOIN
                pg_catalog.pg_description pgtd
                on pgtd.objoid=st.relid and pgtd.objsubid=0
            LEFT JOIN
              pg_catalog.pg_description pgcd
              on pgcd.objoid=st.relid and pgcd.objsubid=att.attnum
            WHERE att.attnum >=0 and {where_clause_suffix}
            ORDER by cluster, schema, name, col_sort_order;
        """.format(
            cluster_source=cluster_source,
            where_clause_suffix=where_clause_suffix,
        )

    def get_scope(self) -> str:
        return 'extractor.postgres_metadata'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc

from pyhocon import ConfigTree

from databuilder import Scoped


class Task(Scoped):
    """
    A Abstract task that can run an abstract task
    """
    @abc.abstractmethod
    def init(self, conf: ConfigTree) -> None:
        pass

    @abc.abstractmethod
    def run(self) -> None:
        """
        Runs a task
        :return:
        """
        pass

    def get_scope(self) -> str:
        return 'task'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import hashlib
from typing import (
    Iterator, List, Optional,
)

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.query.base import QueryBase
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.user import User as UserMetadata


class QueryMetadata(QueryBase):
    """
    Query model. This creates a Query object as well as relationships
    between the Query and the Table(s) that are used within the query.
    The list of TableMetadata corresponding to the tables used in the
    query must be provided. Optionally, the ID of the user that executed
    the query can be provided as well.

    By default, all tables and users must already exist in the database
    before this QueryMetadata object in order to create the relationships.
    Implementers have the option to set `yield_relation_nodes` = True
    in order to create all user and table nodes on the fly at the time
    that this QueryMetadata is created.
    """
    NODE_LABEL = 'Query'
    KEY_FORMAT = '{sql_hash}'

    # Relation between entity and query
    TABLE_QUERY_RELATION_TYPE = 'HAS_QUERY'
    INVERSE_TABLE_QUERY_RELATION_TYPE = 'QUERY_FOR'

    USER_QUERY_RELATION_TYPE = 'EXECUTED_QUERY'
    INVERSE_USER_QUERY_RELATION_TYPE = 'EXECUTED_BY'

    # Attributes
    SQL = 'sql'
    TABLES = 'tables'

    def __init__(
        self,
        sql: str,
        tables: List[TableMetadata],
        clean_sql: Optional[str] = None,              #
        user: Optional[UserMetadata] = None,          # Key for the user that executed the query
        yield_relation_nodes: bool = False  # Allow creation of related nodes if they do not exist
    ):
        """
        :param sql: Full, raw SQL for a given Query
        :param tables: List of table meteadata objects corresponding to tables in the query
        :param clean_sql: A modified sql that should be used to create the hash if available. This
            may be used if you have a query that is run on a set schedule but the where clause has
            a new date or hour value injected before the query is run. You can "clean" that value
            and pass in a SQL string that corresponds to the underlying query - which would should
            remain the same across executions.
        :param user: The user that executed the query.
        :param yield_relation_nodes: A boolean, indicating whether or not all tables and users
            associated to this query should have nodes created if they do not already exist.
        """
        self.sql = sql
        self.clean_sql = clean_sql
        self.sql_hash = self._get_sql_hash(clean_sql or sql)
        self.tables = tables
        self.table_keys = [tm._get_table_key() for tm in tables]
        self.user = user
        self.yield_relation_nodes = yield_relation_nodes
        self._sql_begin = sql[:25] + '...'
        self._node_iter = self._create_next_node()
        self._relation_iter = self._create_relation_iterator()

    def __repr__(self) -> str:
        return f'QueryMetadata(SQL: {self._sql_begin}, Tables: {self.table_keys})'

    def _get_sql_hash(self, sql: str) -> str:
        """
        Generates a deterministic SQL hash. Attempts to remove any formatting from the
        SQL code where possible.
        """
        sql_no_fmt = self._normalize(sql)
        return hashlib.md5(sql_no_fmt.encode('utf-8')).hexdigest()

    def create_next_node(self) -> Optional[GraphNode]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    @staticmethod
    def get_key(sql_hash: str) -> str:
        return QueryMetadata.KEY_FORMAT.format(sql_hash=sql_hash)

    def get_key_self(self) -> str:
        return QueryMetadata.get_key(self.sql_hash)

    def get_query_relations(self) -> List[GraphRelationship]:
        relations = []
        for table_key in self.table_keys:
            table_relation = GraphRelationship(
                start_label=TableMetadata.TABLE_NODE_LABEL,
                end_label=self.NODE_LABEL,
                start_key=table_key,
                end_key=self.get_key_self(),
                type=self.TABLE_QUERY_RELATION_TYPE,
                reverse_type=self.INVERSE_TABLE_QUERY_RELATION_TYPE,
                attributes={}
            )
            relations.append(table_relation)

        if self.user:
            user_relation = GraphRelationship(
                start_label=UserMetadata.USER_NODE_LABEL,
                end_label=self.NODE_LABEL,
                start_key=self.user.get_user_model_key(email=self.user.email),
                end_key=self.get_key_self(),
                type=self.USER_QUERY_RELATION_TYPE,
                reverse_type=self.INVERSE_USER_QUERY_RELATION_TYPE,
                attributes={}
            )
            relations.append(user_relation)
        return relations

    def _create_next_node(self) -> Iterator[GraphNode]:
        """
        Create query nodes
        :return:
        """
        yield GraphNode(
            key=self.get_key_self(),
            label=self.NODE_LABEL,
            attributes={
                self.SQL: self.sql
            }
        )
        if self.yield_relation_nodes:
            for table in self.tables:
                for tbl_item in table._create_next_node():
                    yield tbl_item
            if self.user:
                usr = self.user.create_next_node()
                while usr:
                    yield usr
                    usr = self.user.create_next_node()

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        relations = self.get_query_relations()
        for relation in relations:
            yield relation

        if self.yield_relation_nodes:
            for table in self.tables:
                for tbl_rel in table._create_next_relation():
                    yield tbl_rel
            if self.user:
                for usr_rel in self.user._create_relation_iterator():
                    yield usr_rel



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import re
from collections import namedtuple
from datetime import (
    datetime, timedelta, timezone,
)
from time import sleep
from typing import (
    Any, Dict, Iterator, List, Optional, Tuple,
)

from pyhocon import ConfigTree

from databuilder.extractor.base_bigquery_extractor import BaseBigQueryExtractor

TableColumnUsageTuple = namedtuple('TableColumnUsageTuple', ['database', 'cluster', 'schema',
                                                             'table', 'column', 'email'])

LOGGER = logging.getLogger(__name__)


class BigQueryTableUsageExtractor(BaseBigQueryExtractor):
    """
    An aggregate extractor for bigquery table usage. This class takes the data from
    the stackdriver logging API by filtering on timestamp, bigquery_resource and looking
    for referencedTables in the response.
    """
    TIMESTAMP_KEY = 'timestamp'
    _DEFAULT_SCOPES = ['https://www.googleapis.com/auth/cloud-platform']
    EMAIL_PATTERN = 'email_pattern'
    DELAY_TIME = 'delay_time'
    TABLE_DECORATORS = ['$', '@']
    COUNT_READS_ONLY_FROM_PROJECT_ID_KEY = 'count_reads_only_from_project_id_key'

    def init(self, conf: ConfigTree) -> None:
        BaseBigQueryExtractor.init(self, conf)
        self.timestamp = conf.get_string(
            BigQueryTableUsageExtractor.TIMESTAMP_KEY,
            (datetime.now(timezone.utc) - timedelta(days=1)).strftime(BigQueryTableUsageExtractor.DATE_TIME_FORMAT))

        self.email_pattern = conf.get_string(BigQueryTableUsageExtractor.EMAIL_PATTERN, None)
        self.delay_time = conf.get_int(BigQueryTableUsageExtractor.DELAY_TIME, 100)
        self.table_usage_counts: Dict[TableColumnUsageTuple, int] = {}
        # GCP console allows running queries using tables from a project different from the one the extractor is
        # used for; only usage metadata of referenced tables present in the given project_id_key for the
        # extractor is taken into account and usage metadata of referenced tables from other projects
        # is ignored by "default".
        self.count_reads_only_from_same_project = conf.get_bool(
            BigQueryTableUsageExtractor.COUNT_READS_ONLY_FROM_PROJECT_ID_KEY, True)
        self._count_usage()
        self.iter = iter(self.table_usage_counts)

    def _count_usage(self) -> None:  # noqa: C901
        count = 0
        for entry in self._retrieve_records():
            count += 1
            if count % self.pagesize == 0:
                LOGGER.info(f'Aggregated {count} records')

            if entry is None:
                continue

            try:
                job = entry['protoPayload']['serviceData']['jobCompletedEvent']['job']
            except Exception:
                continue
            if job['jobStatus']['state'] != 'DONE':
                # This job seems not to have finished yet, so we ignore it.
                continue
            if len(job['jobStatus'].get('error', {})) > 0:
                # This job has errors, so we ignore it
                continue

            email = entry['protoPayload']['authenticationInfo']['principalEmail']
            # Query results can be cached and if the source tables remain untouched,
            # bigquery will return it from a 24 hour cache result instead. In that
            # case, referencedTables has been observed to be empty:
            # https://cloud.google.com/logging/docs/reference/audit/bigquery/rest/Shared.Types/AuditData#JobStatistics

            refTables = job['jobStatistics'].get('referencedTables', None)
            if refTables:
                if 'totalTablesProcessed' in job['jobStatistics']:
                    self._create_records(
                        refTables,
                        job['jobStatistics']['totalTablesProcessed'], email,
                        job['jobName']['jobId'])

            refViews = job['jobStatistics'].get('referencedViews', None)
            if refViews:
                if 'totalViewsProcessed' in job['jobStatistics']:
                    self._create_records(
                        refViews, job['jobStatistics']['totalViewsProcessed'],
                        email, job['jobName']['jobId'])

    def _create_records(self, refResources: List[dict], resourcesProcessed: int, email: str,
                        jobId: str) -> None:
        # if email filter is provided, only the email matched with filter will be recorded.
        if self.email_pattern:
            if not re.match(self.email_pattern, email):
                # the usage account not match email pattern
                return

        if len(refResources) != resourcesProcessed:
            LOGGER.warning(f'The number of tables listed in job {jobId} is not consistent')
            return

        for refResource in refResources:
            tableId = refResource.get('tableId')
            datasetId = refResource.get('datasetId')

            if not datasetId or not tableId:
                # handling case when the referenced table is an external table
                # Which doesn't have a datasetId
                continue

            if self._is_anonymous_dataset(datasetId) or self._is_wildcard_table(tableId):
                continue

            tableId = self._remove_table_decorators(tableId)

            if self._is_sharded_table(tableId):
                # Use the prefix of the sharded table as tableId
                tableId = tableId[:-len(self._get_sharded_table_suffix(tableId))]

            if refResource['projectId'] != self.project_id and self.count_reads_only_from_same_project:
                LOGGER.debug(
                    f'Not counting usage for {refResource} since {tableId} '
                    f'is not present in {self.project_id} '
                    f'and {BigQueryTableUsageExtractor.COUNT_READS_ONLY_FROM_PROJECT_ID_KEY} is True')
                continue
            else:
                key = TableColumnUsageTuple(database='bigquery',
                                            cluster=refResource['projectId'],
                                            schema=datasetId,
                                            table=tableId,
                                            column='*',
                                            email=email)

            new_count = self.table_usage_counts.get(key, 0) + 1
            self.table_usage_counts[key] = new_count

    def _retrieve_records(self) -> Iterator[Optional[Dict]]:
        """
        Extracts bigquery log data by looking at the principalEmail in the authenticationInfo block and
        referencedTables in the jobStatistics and filters out log entries of metadata queries.
        :return: Provides a record or None if no more to extract
        """
        body = {
            'resourceNames': [f'projects/{self.project_id}'],
            'pageSize': self.pagesize,
            'filter': 'protoPayload.methodName="jobservice.jobcompleted" AND '
                      'resource.type="bigquery_resource" AND '
                      'NOT protoPayload.serviceData.jobCompletedEvent.job.jobConfiguration.query.query:('
                      'INFORMATION_SCHEMA OR __TABLES__) AND '
                      f'timestamp >= "{self.timestamp}" AND timestamp < "{self.cutoff_time}"'
        }
        for page in self._page_over_results(body):
            for entry in page['entries']:
                yield entry

    def extract(self) -> Optional[Tuple[Any, int]]:
        try:
            key = next(self.iter)
            return key, self.table_usage_counts[key]
        except StopIteration:
            return None

    def _page_over_results(self, body: Dict) -> Iterator[Dict]:
        response = self.logging_service.entries().list(body=body).execute(
            num_retries=BigQueryTableUsageExtractor.NUM_RETRIES)
        while response:
            if 'entries' in response:
                yield response

            try:
                if 'nextPageToken' in response:
                    body['pageToken'] = response['nextPageToken']
                    response = self.logging_service.entries().list(body=body).execute(
                        num_retries=BigQueryTableUsageExtractor.NUM_RETRIES)
                else:
                    response = None
            except Exception:
                # Add a delay when BQ quota exceeds limitation
                sleep(self.delay_time)

    def _remove_table_decorators(self, tableId: str) -> Optional[str]:
        for decorator in BigQueryTableUsageExtractor.TABLE_DECORATORS:
            tableId = tableId.split(decorator)[0]
        return tableId

    def _is_anonymous_dataset(self, datasetId: str) -> bool:
        # temporary/cached results tables are stored in anonymous datasets that have names starting with '_'
        return datasetId.startswith('_')

    def _is_wildcard_table(self, tableId: str) -> bool:
        return '*' in tableId

    def get_scope(self) -> str:
        return 'extractor.bigquery_table_usage'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Iterator, Union

from amundsen_rds.models import RDSModel
from amundsen_rds.models.table import TableTimestamp as RDSTableTimestamp

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.table_serializable import TableSerializable
from databuilder.models.timestamp import timestamp_constants


class TableLastUpdated(GraphSerializable, TableSerializable, AtlasSerializable):
    # constants
    LAST_UPDATED_NODE_LABEL = timestamp_constants.NODE_LABEL
    LAST_UPDATED_KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}/timestamp'
    TIMESTAMP_PROPERTY = timestamp_constants.DEPRECATED_TIMESTAMP_PROPERTY
    TIMESTAMP_NAME_PROPERTY = timestamp_constants.TIMESTAMP_NAME_PROPERTY

    TABLE_LASTUPDATED_RELATION_TYPE = timestamp_constants.LASTUPDATED_RELATION_TYPE
    LASTUPDATED_TABLE_RELATION_TYPE = timestamp_constants.LASTUPDATED_REVERSE_RELATION_TYPE

    def __init__(self,
                 table_name: str,
                 last_updated_time_epoch: int,
                 schema: str,
                 db: str = 'hive',
                 cluster: str = 'gold'
                 ) -> None:
        self.table_name = table_name
        self.last_updated_time = int(last_updated_time_epoch)
        self.schema = schema
        self.db = db
        self.cluster = cluster

        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()

    def __repr__(self) -> str:
        return f"TableLastUpdated(table_name={self.table_name!r}, last_updated_time={self.last_updated_time!r}, " \
               f"schema={self.schema!r}, db={self.db!r}, cluster={self.cluster!r})"

    def create_next_node(self) -> Union[GraphNode, None]:
        # creates new node
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def get_table_model_key(self) -> str:
        # returns formatted string for table name
        return TableMetadata.TABLE_KEY_FORMAT.format(db=self.db,
                                                     cluster=self.cluster,
                                                     schema=self.schema,
                                                     tbl=self.table_name)

    def get_last_updated_model_key(self) -> str:
        # returns formatted string for last updated name
        return TableLastUpdated.LAST_UPDATED_KEY_FORMAT.format(db=self.db,
                                                               cluster=self.cluster,
                                                               schema=self.schema,
                                                               tbl=self.table_name)

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create a last_updated node
        :return:
        """
        node = GraphNode(
            key=self.get_last_updated_model_key(),
            label=TableLastUpdated.LAST_UPDATED_NODE_LABEL,
            attributes={
                TableLastUpdated.TIMESTAMP_PROPERTY: self.last_updated_time,
                timestamp_constants.TIMESTAMP_PROPERTY: self.last_updated_time,
                TableLastUpdated.TIMESTAMP_NAME_PROPERTY: timestamp_constants.TimestampName.last_updated_timestamp.name
            }
        )
        yield node

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        """
        Create relations mapping last updated node with table node
        :return:
        """
        relationship = GraphRelationship(
            start_label=TableMetadata.TABLE_NODE_LABEL,
            start_key=self.get_table_model_key(),
            end_label=TableLastUpdated.LAST_UPDATED_NODE_LABEL,
            end_key=self.get_last_updated_model_key(),
            type=TableLastUpdated.TABLE_LASTUPDATED_RELATION_TYPE,
            reverse_type=TableLastUpdated.LASTUPDATED_TABLE_RELATION_TYPE,
            attributes={}
        )
        yield relationship

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        """
        Create a table timestamp record
        :return:
        """
        record = RDSTableTimestamp(
            rk=self.get_last_updated_model_key(),
            last_updated_timestamp=self.last_updated_time,
            timestamp=self.last_updated_time,
            name=timestamp_constants.TimestampName.last_updated_timestamp.name,
            table_rk=self.get_table_model_key()
        )
        yield record

    # Atlas automatically updates `updateTime` of an entity if it's changed (along with storing audit info what changed)
    # so we don't really need to implement those methods. The reason they exist at all is so loader class doesn't fail
    # if extractor extracts this info.
    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        pass

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        pass



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import csv
import logging
from typing import Any

from pyhocon import ConfigTree

from databuilder.loader.base_loader import Loader

LOGGER = logging.getLogger(__name__)


class FileSystemCSVLoader(Loader):
    """
    Loader class to write csv files to Local FileSystem
    """

    def init(self, conf: ConfigTree) -> None:
        """
        Initialize file handlers from conf
        :param conf:
        """
        self.conf = conf
        self.file_path = self.conf.get_string('file_path')
        self.file_mode = self.conf.get_string('mode', 'w')

        self.file_handler = open(self.file_path, self.file_mode)

    def load(self, record: Any) -> None:
        """
        Write record object as csv to file
        :param record:
        :return:
        """
        if not record:
            return

        if not hasattr(self, 'writer'):
            self.writer = csv.DictWriter(self.file_handler,
                                         fieldnames=vars(record).keys())
            self.writer.writeheader()

        self.writer.writerow(vars(record))
        self.file_handler.flush()

    def close(self) -> None:
        """
        Close file handlers
        :return:
        """
        try:
            if self.file_handler:
                self.file_handler.close()
        except Exception as e:
            LOGGER.warning("Failed trying to close a file handler! %s", e)

    def get_scope(self) -> str:
        return "loader.filesystem.csv"



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import textwrap

# These queries are meant to be used to extract search metadata from neo4j
# using SearchMetadatatoElasticasearchTask

NEO4J_TABLE_CYPHER_QUERY = textwrap.dedent(
    """
    MATCH (db:Database)<-[:CLUSTER_OF]-(cluster:Cluster)
    <-[:SCHEMA_OF]-(schema:Schema)<-[:TABLE_OF]-(table:Table)
    {publish_tag_filter}
    OPTIONAL MATCH (schema)-[:DESCRIPTION]->(schema_description:Description)
    OPTIONAL MATCH (table)-[:DESCRIPTION]->(table_description:Description)
    OPTIONAL MATCH (table)-[:TAGGED_BY]->(tags:Tag) WHERE tags.tag_type='default'
    WITH db, cluster, schema, schema_description, table, table_description,
    COLLECT(DISTINCT toLower(tags.key)) as tags
    OPTIONAL MATCH (table)-[:HAS_BADGE]->(badges:Badge)
    WITH db, cluster, schema, schema_description, table, table_description, tags,
    COLLECT(DISTINCT toLower(badges.key)) as badges
    MATCH (table)-[:COLUMN]->(col:Column)
    OPTIONAL MATCH (col)-[:DESCRIPTION]->(col_description:Description)
    WITH db, cluster, schema, schema_description, table, table_description, tags, badges,
    COLLECT(toLower(col.name)) AS columns, COLLECT(col_description.description) AS column_descriptions
    OPTIONAL MATCH (table)-[:LAST_UPDATED_AT]->(time_stamp:Timestamp)
    {additional_field_match}
    RETURN toLower(db.name) as database, cluster.name AS cluster, toLower(schema.name) AS schema,
    schema_description.description AS schema_description,
    toLower(table.name) AS name, table.key AS key, table_description.description AS description,
    time_stamp.last_updated_timestamp AS last_updated_timestamp,
    {{
        {usage_fields}
    }} AS usage,
    columns,
    column_descriptions,
    tags,
    {additional_field_return}
    badges
    ORDER BY table.name;
    """
)

DEFAULT_TABLE_QUERY = NEO4J_TABLE_CYPHER_QUERY.format(
    publish_tag_filter='',
    additional_field_match="""
        OPTIONAL MATCH (table)-[read:READ_BY]->(user:User)
        WITH db, cluster, schema, schema_description, table, table_description, time_stamp,
        tags, badges, columns, column_descriptions, SUM(read.read_count) AS total_usage,
        COUNT(DISTINCT user.email) AS unique_usage
    """,
    usage_fields="""
        total_usage: CASE total_usage
        WHEN 0 THEN null
        ELSE total_usage
        END,
        unique_usage: CASE unique_usage
        WHEN 0 THEN null
        ELSE unique_usage
        END
    """,
    additional_field_return='')

NEO4J_DASHBOARD_CYPHER_QUERY = textwrap.dedent(
    """
        MATCH (dashboard:Dashboard)-[:DASHBOARD_OF]->(dbg:Dashboardgroup)
        -[:DASHBOARD_GROUP_OF]->(cluster:Cluster)
        {publish_tag_filter}
        OPTIONAL MATCH (dbg)-[:DESCRIPTION]->(dbg_descr:Description)
        OPTIONAL MATCH (dashboard)-[:DESCRIPTION]->(db_descr:Description)
        OPTIONAL MATCH (dashboard)-[:EXECUTED]->(last_exec:Execution)
        WHERE split(last_exec.key, '/')[5] = '_last_successful_execution'
        OPTIONAL MATCH (dashboard)-[:HAS_QUERY]->(query:Query)-[:HAS_CHART]->(chart:Chart)
        WITH dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, COLLECT(DISTINCT query.name) as query_names,
        COLLECT(DISTINCT chart.name) as chart_names
        OPTIONAL MATCH (dashboard)-[:TAGGED_BY]->(tags:Tag) WHERE tags.tag_type='default'
        WITH dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, query_names, chart_names,
        COLLECT(DISTINCT toLower(tags.key)) as tags
        OPTIONAL MATCH (dashboard)-[:HAS_BADGE]->(badges:Badge)
        WITH  dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, query_names, chart_names, tags,
        COLLECT(DISTINCT badges.key) as badges
        {additional_field_match}
        RETURN toLower(dbg.name) as group_name, toLower(dashboard.name) as name, cluster.name as cluster,
        {additional_field_return}
        {{
            {usage_fields}
        }} AS usage,
        coalesce(db_descr.description, '') as description,
        coalesce(dbg.description, '') as group_description, dbg.dashboard_group_url as group_url,
        dashboard.dashboard_url as url, dashboard.key as key, dashboard.key as uri,
        split(dashboard.key, '_')[0] as product, toInteger(last_exec.timestamp) as last_successful_run_timestamp,
        query_names, chart_names, tags, badges
        order by dbg.name
    """
)

DEFAULT_DASHBOARD_QUERY = NEO4J_DASHBOARD_CYPHER_QUERY.format(
    publish_tag_filter='',
    additional_field_match="""
        OPTIONAL MATCH (dashboard)-[read:READ_BY]->(user:User)
        WITH  dashboard, dbg, db_descr, dbg_descr, cluster, last_exec, query_names, chart_names, tags,
        badges, SUM(read.read_count) AS total_usage
    """,
    usage_fields="""
        total_usage: CASE total_usage
        WHEN 0 THEN null
        ELSE total_usage
        END
    """,
    additional_field_return='')

NEO4J_USER_CYPHER_QUERY = textwrap.dedent(
    """
    MATCH (user:User)
    {additional_field_match}
    OPTIONAL MATCH (user)-[read:READ]->(a)
    OPTIONAL MATCH (user)-[own:OWNER_OF]->(b)
    OPTIONAL MATCH (user)-[follow:FOLLOWED_BY]->(c)
    OPTIONAL MATCH (user)-[manage_by:MANAGE_BY]->(manager)
    {publish_tag_filter}
    with user, a, b, c, read, own, follow, manager
    where user.full_name is not null
    return user.email as key, user.first_name as first_name, user.last_name as last_name,
    {{
        {usage_fields}
    }} AS usage,
    user.full_name as name, user.github_username as github_username, user.team_name as team_name,
    user.employee_type as employee_type, manager.email as manager_email,
    {additional_field_return}
    user.slack_id as slack_id, toBoolean(user.is_active) as is_active, user.role_name as role_name
    order by user.email
    """
)

DEFAULT_USER_QUERY = NEO4J_USER_CYPHER_QUERY.format(
    publish_tag_filter='',
    additional_field_match='',
    usage_fields="""
        total_read: CASE sum(read.read_count)
        WHEN 0 THEN null
        ELSE sum(read.read_count)
        END,
        total_own: CASE count(distinct b)
        WHEN 0 THEN null
        ELSE count(distinct b)
        END,
        total_follow: CASE count(distinct c)
        WHEN 0 THEN null
        ELSE count(distinct c)
        END
    """,
    additional_field_return='')

NEO4J_FEATURE_CYPHER_QUERY = textwrap.dedent(
    """
        MATCH (feature:Feature)
        {publish_tag_filter}
        OPTIONAL MATCH (fg:Feature_Group)-[:GROUPS]->(feature)
        OPTIONAL MATCH (db:Database)-[:AVAILABLE_FEATURE]->(feature)
        OPTIONAL MATCH (feature)-[:DESCRIPTION]->(desc:Description)
        OPTIONAL MATCH (feature)-[:TAGGED_BY]->(tag:Tag)
        OPTIONAL MATCH (feature)-[:HAS_BADGE]->(badge:Badge)
        {additional_field_match}
        RETURN
        toLower(fg.name) as feature_group,
        toLower(feature.name) as name,
        feature.version as version,
        feature.key as key,
        {additional_field_return}
        {{
            {usage_fields}
        }} AS usage,
        feature.status as status,
        toLower(feature.entity) as entity,
        desc.description as description,
        db.name as availability,
        COLLECT(DISTINCT toLower(badge.key)) as badges,
        COLLECT(DISTINCT toLower(tag.key)) as tags,
        toInteger(feature.last_updated_timestamp) as last_updated_timestamp
        order by feature_group, name, version
    """
)

DEFAULT_FEATURE_QUERY = NEO4J_FEATURE_CYPHER_QUERY.format(
    publish_tag_filter='',
    additional_field_match="""
        OPTIONAL MATCH (feature)-[read:READ_BY]->(user:User)
        WITH feature, fg, db, desc, tag, badge, SUM(read.read_count) as total_usage
    """,
    usage_fields="""
        total_usage: CASE total_usage
        WHEN 0 THEN null
        ELSE total_usage
        END
        """,
    additional_field_return='')



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import copy
from typing import (
    TYPE_CHECKING, Any, Dict, Iterable, Iterator, List, Optional, Set, Union,
)

from amundsen_common.utils.atlas import (
    AtlasCommonParams, AtlasCommonTypes, AtlasTableTypes,
)
from amundsen_rds.models import RDSModel
from amundsen_rds.models.cluster import Cluster as RDSCluster
from amundsen_rds.models.column import (
    ColumnBadge as RDSColumnBadge, ColumnDescription as RDSColumnDescription, TableColumn as RDSTableColumn,
)
from amundsen_rds.models.database import Database as RDSDatabase
from amundsen_rds.models.schema import Schema as RDSSchema
from amundsen_rds.models.table import (
    Table as RDSTable, TableDescription as RDSTableDescription,
    TableProgrammaticDescription as RDSTableProgrammaticDescription, TableTag as RDSTableTag,
)
from amundsen_rds.models.tag import Tag as RDSTag

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.badge import Badge, BadgeMetadata
from databuilder.models.cluster import cluster_constants
from databuilder.models.description_metadata import (  # noqa: F401
    DESCRIPTION_NODE_LABEL, DESCRIPTION_NODE_LABEL_VAL, DescriptionMetadata,
)
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.schema import schema_constant
from databuilder.models.table_serializable import TableSerializable

if TYPE_CHECKING:
    from databuilder.models.type_metadata import TypeMetadata

from databuilder.serializers.atlas_serializer import (
    add_entity_relationship, get_entity_attrs, get_entity_relationships,
)
from databuilder.utils.atlas import AtlasRelationshipTypes, AtlasSerializedEntityOperation


def _format_as_list(tags: Union[List, str, None]) -> List:
    if tags is None:
        tags = []
    if isinstance(tags, str):
        tags = list(filter(None, tags.split(',')))
    if isinstance(tags, list):
        tags = [tag.lower().strip() for tag in tags]
    return tags


class TagMetadata(GraphSerializable, TableSerializable, AtlasSerializable):
    TAG_NODE_LABEL = 'Tag'
    TAG_KEY_FORMAT = '{tag}'
    TAG_TYPE = 'tag_type'
    DEFAULT_TYPE = 'default'
    BADGE_TYPE = 'badge'
    DASHBOARD_TYPE = 'dashboard'
    METRIC_TYPE = 'metric'

    TAG_ENTITY_RELATION_TYPE = 'TAG'
    ENTITY_TAG_RELATION_TYPE = 'TAGGED_BY'

    def __init__(self,
                 name: str,
                 tag_type: str = 'default',
                 ):
        self._name = name
        self._tag_type = tag_type
        self._nodes = self._create_node_iterator()
        self._relations = self._create_relation_iterator()
        self._records = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()

    @staticmethod
    def get_tag_key(name: str) -> str:
        if not name:
            return ''
        return TagMetadata.TAG_KEY_FORMAT.format(tag=name)

    def get_node(self) -> GraphNode:
        node = GraphNode(
            key=TagMetadata.get_tag_key(self._name),
            label=TagMetadata.TAG_NODE_LABEL,
            attributes={
                TagMetadata.TAG_TYPE: self._tag_type
            }
        )
        return node

    def get_record(self) -> RDSModel:
        record = RDSTag(
            rk=TagMetadata.get_tag_key(self._name),
            tag_type=self._tag_type
        )
        return record

    def create_next_node(self) -> Optional[GraphNode]:
        # return the string representation of the data
        try:
            return next(self._nodes)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        # We don't emit any relations for Tag ingestion
        try:
            return next(self._relations)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._records)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        node = self.get_node()
        yield node

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        return
        yield

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        record = self.get_record()
        yield record

    def _create_atlas_glossary_entity(self) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._name),
            ('glossary', self._tag_type),
            ('term', self._name)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        entity = AtlasEntity(
            typeName=AtlasCommonTypes.tag,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=None
        )

        return entity

    def create_atlas_tag_relation(self, table_key: str) -> AtlasRelationship:
        table_relationship = AtlasRelationship(
            relationshipType=AtlasRelationshipTypes.tag,
            entityType1=AtlasCommonTypes.data_set,
            entityQualifiedName1=table_key,
            entityType2=AtlasRelationshipTypes.tag,
            entityQualifiedName2=f'glossary={self._tag_type},term={self._name}',
            attributes={}
        )

        return table_relationship

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        pass

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        yield self._create_atlas_glossary_entity()

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None


class ColumnMetadata:
    COLUMN_NODE_LABEL = 'Column'
    COLUMN_KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}'
    COLUMN_NAME = 'name'
    COLUMN_TYPE = 'col_type'
    COLUMN_ORDER = 'sort_order'
    COLUMN_DESCRIPTION = 'description'
    COLUMN_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}/{description_id}'

    def __init__(self,
                 name: str,
                 description: Union[str, None],
                 col_type: str,
                 sort_order: int,
                 badges: Union[List[str], None] = None,
                 ) -> None:
        """
        TODO: Add stats
        :param name:
        :param description:
        :param col_type:
        :param sort_order:
        :param badges: Optional. Column level badges
        """
        self.name = name
        self.description = DescriptionMetadata.create_description_metadata(source=None,
                                                                           text=description)
        self.type = col_type
        self.sort_order = sort_order
        formatted_badges = _format_as_list(badges)
        self.badges = [Badge(badge, 'column') for badge in formatted_badges]

        # The following fields are populated by the ComplexTypeTransformer
        self._column_key: Optional[str] = None
        self._type_metadata: Optional[TypeMetadata] = None

    def __repr__(self) -> str:
        return f'ColumnMetadata({self.name!r}, {self.description!r}, {self.type!r}, ' \
               f'{self.sort_order!r}, {self.badges!r})'

    def get_column_key(self) -> Optional[str]:
        return self._column_key

    def set_column_key(self, col_key: str) -> None:
        self._column_key = col_key

    def get_type_metadata(self) -> Optional['TypeMetadata']:
        return self._type_metadata

    def set_type_metadata(self, type_metadata: 'TypeMetadata') -> None:
        self._type_metadata = type_metadata


class TableMetadata(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    Table metadata that contains columns. It implements Neo4jCsvSerializable so that it can be serialized to produce
    Table, Column and relation of those along with relationship with table and schema. Additionally, it will create
    Database, Cluster, and Schema with relastionships between those.
    These are being created here as it does not make much sense to have different extraction to produce this. As
    database, cluster, schema would be very repititive with low cardinality, it will perform de-dupe so that publisher
    won't need to publish same nodes, relationships.

    This class can be used for both table and view metadata. If it is a View, is_view=True should be passed in.
    """
    TABLE_NODE_LABEL = 'Table'
    TABLE_KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}'
    TABLE_NAME = 'name'
    IS_VIEW = 'is_view'

    TABLE_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{description_id}'

    DATABASE_NODE_LABEL = 'Database'
    DATABASE_KEY_FORMAT = 'database://{db}'
    DATABASE_CLUSTER_RELATION_TYPE = cluster_constants.CLUSTER_RELATION_TYPE
    CLUSTER_DATABASE_RELATION_TYPE = cluster_constants.CLUSTER_REVERSE_RELATION_TYPE

    CLUSTER_NODE_LABEL = cluster_constants.CLUSTER_NODE_LABEL
    CLUSTER_KEY_FORMAT = '{db}://{cluster}'
    CLUSTER_SCHEMA_RELATION_TYPE = schema_constant.SCHEMA_RELATION_TYPE
    SCHEMA_CLUSTER_RELATION_TYPE = schema_constant.SCHEMA_REVERSE_RELATION_TYPE

    SCHEMA_NODE_LABEL = schema_constant.SCHEMA_NODE_LABEL
    SCHEMA_KEY_FORMAT = schema_constant.DATABASE_SCHEMA_KEY_FORMAT
    SCHEMA_TABLE_RELATION_TYPE = 'TABLE'
    TABLE_SCHEMA_RELATION_TYPE = 'TABLE_OF'

    TABLE_COL_RELATION_TYPE = 'COLUMN'
    COL_TABLE_RELATION_TYPE = 'COLUMN_OF'

    TABLE_TAG_RELATION_TYPE = TagMetadata.ENTITY_TAG_RELATION_TYPE
    TAG_TABLE_RELATION_TYPE = TagMetadata.TAG_ENTITY_RELATION_TYPE

    # Only for deduping database, cluster, and schema (table and column will be always processed)
    serialized_nodes_keys: Set[Any] = set()
    serialized_rels_keys: Set[Any] = set()
    serialized_records_keys: Set[Any] = set()

    def __init__(self,
                 database: str,
                 cluster: str,
                 schema: str,
                 name: str,
                 description: Union[str, None],
                 columns: Iterable[ColumnMetadata] = None,
                 is_view: bool = False,
                 tags: Union[List, str] = None,
                 description_source: Union[str, None] = None,
                 **kwargs: Any
                 ) -> None:
        """
        :param database:
        :param cluster:
        :param schema:
        :param name:
        :param description:
        :param columns:
        :param is_view: Indicate whether the table is a view or not
        :param tags:
        :param description_source: Optional. Where the description is coming from. Used to compose unique id.
        :param kwargs: Put additional attributes to the table model if there is any.
        """
        self.database = database
        self.cluster = cluster
        self.schema = schema
        self.name = name
        self.description = DescriptionMetadata.create_description_metadata(text=description, source=description_source)
        self.columns = columns if columns else []
        self.is_view = is_view
        self.attrs: Optional[Dict[str, Any]] = None

        self.tags = _format_as_list(tags)

        if kwargs:
            self.attrs = copy.deepcopy(kwargs)

        self._node_iterator = self._create_next_node()
        self._relation_iterator = self._create_next_relation()
        self._record_iterator = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()
        self._atlas_relation_iterator = self._create_atlas_relation_iterator()

    def __repr__(self) -> str:
        return f'TableMetadata({self.database!r}, {self.cluster!r}, {self.schema!r}, {self.name!r} ' \
               f'{self.description!r}, {self.columns!r}, {self.is_view!r}, {self.tags!r})'

    def _get_table_key(self) -> str:
        return TableMetadata.TABLE_KEY_FORMAT.format(db=self.database,
                                                     cluster=self.cluster,
                                                     schema=self.schema,
                                                     tbl=self.name)

    def _get_table_description_key(self, description: DescriptionMetadata) -> str:
        return TableMetadata.TABLE_DESCRIPTION_FORMAT.format(db=self.database,
                                                             cluster=self.cluster,
                                                             schema=self.schema,
                                                             tbl=self.name,
                                                             description_id=description.get_description_id())

    def _get_database_key(self) -> str:
        return TableMetadata.DATABASE_KEY_FORMAT.format(db=self.database)

    def _get_cluster_key(self) -> str:
        return TableMetadata.CLUSTER_KEY_FORMAT.format(db=self.database,
                                                       cluster=self.cluster)

    def _get_schema_key(self) -> str:
        return TableMetadata.SCHEMA_KEY_FORMAT.format(db=self.database,
                                                      cluster=self.cluster,
                                                      schema=self.schema)

    def _get_col_key(self, col: ColumnMetadata) -> str:
        return ColumnMetadata.COLUMN_KEY_FORMAT.format(db=self.database,
                                                       cluster=self.cluster,
                                                       schema=self.schema,
                                                       tbl=self.name,
                                                       col=col.name,
                                                       badges=col.badges)

    def _get_col_description_key(self,
                                 col: ColumnMetadata,
                                 description: DescriptionMetadata) -> str:
        return ColumnMetadata.COLUMN_DESCRIPTION_FORMAT.format(db=self.database,
                                                               cluster=self.cluster,
                                                               schema=self.schema,
                                                               tbl=self.name,
                                                               col=col.name,
                                                               description_id=description.get_description_id())

    @staticmethod
    def format_tags(tags: Union[List, str, None]) -> List:
        return _format_as_list(tags)

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_next_node(self) -> Iterator[GraphNode]:
        yield self._create_table_node()

        if self.description:
            node_key = self._get_table_description_key(self.description)
            yield self.description.get_node(node_key)

        # Create the table tag nodes
        if self.tags:
            for tag in self.tags:
                tag_node = TagMetadata(tag).get_node()
                yield tag_node

        for col in self.columns:
            yield from self._create_column_nodes(col)

        # Database, cluster, schema
        others = [
            GraphNode(
                key=self._get_database_key(),
                label=TableMetadata.DATABASE_NODE_LABEL,
                attributes={
                    'name': self.database
                }
            ),
            GraphNode(
                key=self._get_cluster_key(),
                label=TableMetadata.CLUSTER_NODE_LABEL,
                attributes={
                    'name': self.cluster
                }
            ),
            GraphNode(
                key=self._get_schema_key(),
                label=TableMetadata.SCHEMA_NODE_LABEL,
                attributes={
                    'name': self.schema
                }
            )
        ]

        for node_tuple in others:
            if node_tuple.key not in TableMetadata.serialized_nodes_keys:
                TableMetadata.serialized_nodes_keys.add(node_tuple.key)
                yield node_tuple

    def _create_table_node(self) -> GraphNode:
        table_attributes = {
            TableMetadata.TABLE_NAME: self.name,
            TableMetadata.IS_VIEW: self.is_view
        }
        if self.attrs:
            for k, v in self.attrs.items():
                if k not in table_attributes:
                    table_attributes[k] = v

        return GraphNode(
            key=self._get_table_key(),
            label=TableMetadata.TABLE_NODE_LABEL,
            attributes=table_attributes
        )

    def _create_column_nodes(self, col: ColumnMetadata) -> Iterator[GraphNode]:
        column_node = GraphNode(
            key=self._get_col_key(col),
            label=ColumnMetadata.COLUMN_NODE_LABEL,
            attributes={
                ColumnMetadata.COLUMN_NAME: col.name,
                ColumnMetadata.COLUMN_TYPE: col.type,
                ColumnMetadata.COLUMN_ORDER: col.sort_order
            }
        )
        yield column_node

        if col.description:
            node_key = self._get_col_description_key(col, col.description)
            yield col.description.get_node(node_key)

        if col.badges:
            col_badge_metadata = BadgeMetadata(
                start_label=ColumnMetadata.COLUMN_NODE_LABEL,
                start_key=self._get_col_key(col),
                badges=col.badges)
            badge_nodes = col_badge_metadata.get_badge_nodes()
            for node in badge_nodes:
                yield node

        type_metadata = col.get_type_metadata()
        if type_metadata:
            yield from type_metadata.create_node_iterator()

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_next_relation(self) -> Iterator[GraphRelationship]:
        schema_table_relationship = GraphRelationship(
            start_key=self._get_schema_key(),
            start_label=TableMetadata.SCHEMA_NODE_LABEL,
            end_key=self._get_table_key(),
            end_label=TableMetadata.TABLE_NODE_LABEL,
            type=TableMetadata.SCHEMA_TABLE_RELATION_TYPE,
            reverse_type=TableMetadata.TABLE_SCHEMA_RELATION_TYPE,
            attributes={}
        )
        yield schema_table_relationship

        if self.description:
            yield self.description.get_relation(TableMetadata.TABLE_NODE_LABEL,
                                                self._get_table_key(),
                                                self._get_table_description_key(self.description))

        if self.tags:
            for tag in self.tags:
                tag_relationship = GraphRelationship(
                    start_label=TableMetadata.TABLE_NODE_LABEL,
                    start_key=self._get_table_key(),
                    end_label=TagMetadata.TAG_NODE_LABEL,
                    end_key=TagMetadata.get_tag_key(tag),
                    type=TableMetadata.TABLE_TAG_RELATION_TYPE,
                    reverse_type=TableMetadata.TAG_TABLE_RELATION_TYPE,
                    attributes={}
                )
                yield tag_relationship

        for col in self.columns:
            yield from self._create_column_relations(col)

        others = [
            GraphRelationship(
                start_label=TableMetadata.DATABASE_NODE_LABEL,
                end_label=TableMetadata.CLUSTER_NODE_LABEL,
                start_key=self._get_database_key(),
                end_key=self._get_cluster_key(),
                type=TableMetadata.DATABASE_CLUSTER_RELATION_TYPE,
                reverse_type=TableMetadata.CLUSTER_DATABASE_RELATION_TYPE,
                attributes={}
            ),
            GraphRelationship(
                start_label=TableMetadata.CLUSTER_NODE_LABEL,
                end_label=TableMetadata.SCHEMA_NODE_LABEL,
                start_key=self._get_cluster_key(),
                end_key=self._get_schema_key(),
                type=TableMetadata.CLUSTER_SCHEMA_RELATION_TYPE,
                reverse_type=TableMetadata.SCHEMA_CLUSTER_RELATION_TYPE,
                attributes={}
            )
        ]

        for rel_tuple in others:
            if (rel_tuple.start_key, rel_tuple.end_key, rel_tuple.type) not in TableMetadata.serialized_rels_keys:
                TableMetadata.serialized_rels_keys.add((rel_tuple.start_key, rel_tuple.end_key, rel_tuple.type))
                yield rel_tuple

    def _create_column_relations(self, col: ColumnMetadata) -> Iterator[GraphRelationship]:
        column_relationship = GraphRelationship(
            start_label=TableMetadata.TABLE_NODE_LABEL,
            start_key=self._get_table_key(),
            end_label=ColumnMetadata.COLUMN_NODE_LABEL,
            end_key=self._get_col_key(col),
            type=TableMetadata.TABLE_COL_RELATION_TYPE,
            reverse_type=TableMetadata.COL_TABLE_RELATION_TYPE,
            attributes={}
        )
        yield column_relationship

        if col.description:
            yield col.description.get_relation(
                ColumnMetadata.COLUMN_NODE_LABEL,
                self._get_col_key(col),
                self._get_col_description_key(col, col.description)
            )

        if col.badges:
            badge_metadata = BadgeMetadata(start_label=ColumnMetadata.COLUMN_NODE_LABEL,
                                           start_key=self._get_col_key(col),
                                           badges=col.badges)
            badge_relations = badge_metadata.get_badge_relations()
            for relation in badge_relations:
                yield relation

        type_metadata = col.get_type_metadata()
        if type_metadata:
            yield from type_metadata.create_relation_iterator()

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iterator)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        # Database, Cluster, Schema
        others: List[RDSModel] = [
            RDSDatabase(
                rk=self._get_database_key(),
                name=self.database
            ),
            RDSCluster(
                rk=self._get_cluster_key(),
                name=self.cluster,
                database_rk=self._get_database_key()
            ),
            RDSSchema(
                rk=self._get_schema_key(),
                name=self.schema,
                cluster_rk=self._get_cluster_key()
            )
        ]

        for record in others:
            if record.rk not in TableMetadata.serialized_records_keys:
                TableMetadata.serialized_records_keys.add(record.rk)
                yield record

        # Table
        yield RDSTable(
            rk=self._get_table_key(),
            name=self.name,
            is_view=self.is_view,
            schema_rk=self._get_schema_key()
        )

        # Table description
        if self.description:
            description_record_key = self._get_table_description_key(self.description)
            if self.description.label == DescriptionMetadata.DESCRIPTION_NODE_LABEL:
                yield RDSTableDescription(
                    rk=description_record_key,
                    description_source=self.description.source,
                    description=self.description.text,
                    table_rk=self._get_table_key()
                )
            else:
                yield RDSTableProgrammaticDescription(
                    rk=description_record_key,
                    description_source=self.description.source,
                    description=self.description.text,
                    table_rk=self._get_table_key()
                )

        # Tag
        for tag in self.tags:
            tag_record = TagMetadata(tag).get_record()
            yield tag_record

            table_tag_record = RDSTableTag(
                table_rk=self._get_table_key(),
                tag_rk=TagMetadata.get_tag_key(tag)
            )
            yield table_tag_record

        # Column
        for col in self.columns:
            yield RDSTableColumn(
                rk=self._get_col_key(col),
                name=col.name,
                type=col.type,
                sort_order=col.sort_order,
                table_rk=self._get_table_key()
            )

            if col.description:
                description_record_key = self._get_col_description_key(col, col.description)
                yield RDSColumnDescription(
                    rk=description_record_key,
                    description_source=col.description.source,
                    description=col.description.text,
                    column_rk=self._get_col_key(col)
                )

            if col.badges:
                badge_metadata = BadgeMetadata(
                    start_label=ColumnMetadata.COLUMN_NODE_LABEL,
                    start_key=self._get_col_key(col),
                    badges=col.badges
                )

                badge_records = badge_metadata.get_badge_records()
                for badge_record in badge_records:
                    yield badge_record

                    column_badge_record = RDSColumnBadge(
                        column_rk=self._get_col_key(col),
                        badge_rk=badge_record.rk
                    )
                    yield column_badge_record

    def _create_atlas_cluster_entity(self) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._get_cluster_key()),
            ('name', self.cluster),
            ('displayName', self.cluster)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        entity = AtlasEntity(
            typeName=AtlasCommonTypes.cluster,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=None
        )

        return entity

    def _create_atlas_database_entity(self) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._get_database_key()),
            ('name', self.database),
            ('displayName', self.database)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        relationship_list = []  # type: ignore

        add_entity_relationship(
            relationship_list,
            'cluster',
            AtlasCommonTypes.cluster,
            self._get_cluster_key()
        )

        entity = AtlasEntity(
            typeName=AtlasTableTypes.database,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=get_entity_relationships(relationship_list)
        )

        return entity

    def _create_atlas_schema_entity(self) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._get_schema_key()),
            ('name', self.schema),
            ('displayName', self.schema)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        relationship_list = []  # type: ignore

        add_entity_relationship(
            relationship_list,
            'cluster',
            AtlasCommonTypes.cluster,
            self._get_cluster_key()
        )

        entity = AtlasEntity(
            typeName=AtlasTableTypes.schema,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=get_entity_relationships(relationship_list)
        )

        return entity

    def _create_atlas_table_entity(self) -> AtlasEntity:
        table_type = 'table' if not self.is_view else 'view'

        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._get_table_key()),
            ('name', self.name),
            ('tableType', table_type),
            ('description', self.description.text if self.description else ''),
            ('displayName', self.name)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        relationship_list = []  # type: ignore

        add_entity_relationship(
            relationship_list,
            'amundsen_schema',
            AtlasTableTypes.schema,
            self._get_schema_key()
        )

        entity = AtlasEntity(
            typeName=AtlasTableTypes.table,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=get_entity_relationships(relationship_list)
        )

        return entity

    def _create_atlas_column_entity(self, column_metadata: ColumnMetadata) -> AtlasEntity:
        qualified_name = column_metadata.COLUMN_KEY_FORMAT.format(db=self.database,
                                                                  cluster=self.cluster,
                                                                  schema=self.schema,
                                                                  tbl=self.name,
                                                                  col=column_metadata.name)
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, qualified_name),
            ('name', column_metadata.name or ''),
            ('description', column_metadata.description.text if column_metadata.description else ''),
            ('type', column_metadata.type),
            ('position', column_metadata.sort_order),
            ('displayName', column_metadata.name or '')
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        relationship_list = []  # type: ignore

        add_entity_relationship(
            relationship_list,
            'table',
            AtlasTableTypes.table,
            self._get_table_key()
        )

        entity = AtlasEntity(
            typeName=AtlasTableTypes.column,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=get_entity_relationships(relationship_list)
        )

        return entity

    def _create_next_atlas_relation(self) -> Iterator[AtlasRelationship]:
        pass

    def _create_atlas_relation_iterator(self) -> Iterator[AtlasRelationship]:
        for tag in self.tags:
            tag_relation = TagMetadata(tag).create_atlas_tag_relation(self._get_table_key())
            yield tag_relation

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        try:
            return next(self._atlas_relation_iterator)
        except StopIteration:
            return None

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        yield self._create_atlas_cluster_entity()
        yield self._create_atlas_database_entity()
        yield self._create_atlas_schema_entity()
        yield self._create_atlas_table_entity()

        for col in self.columns:
            yield self._create_atlas_column_entity(col)

        if self.tags:
            for tag in self.tags:
                tag_entity = TagMetadata(tag).create_next_atlas_entity()
                if tag_entity:
                    yield tag_entity

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Iterator, List, Tuple, Union,
)

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasTableTypes
from amundsen_rds.models import RDSModel
from amundsen_rds.models.table import TableWatermark as RDSTableWatermark

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers.atlas_serializer import (
    add_entity_relationship, get_entity_attrs, get_entity_relationships,
)
from databuilder.utils.atlas import AtlasSerializedEntityOperation


class Watermark(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    Table watermark result model.
    Each instance represents one row of table watermark result.
    """
    LABEL = 'Watermark'
    KEY_FORMAT = '{database}://{cluster}.{schema}' \
                 '/{table}/{part_type}/'
    WATERMARK_TABLE_RELATION_TYPE = 'BELONG_TO_TABLE'
    TABLE_WATERMARK_RELATION_TYPE = 'WATERMARK'

    def __init__(self,
                 create_time: str,
                 database: str,
                 schema: str,
                 table_name: str,
                 part_name: str,
                 part_type: str = 'high_watermark',
                 cluster: str = 'gold',
                 ) -> None:
        self.create_time = create_time
        self.database = database
        self.schema = schema
        self.table = table_name
        self.parts: List[Tuple[str, str]] = []

        if '=' not in part_name:
            raise Exception('Only partition table has high watermark')

        # currently we don't consider nested partitions
        idx = part_name.find('=')
        name, value = part_name[:idx], part_name[idx + 1:]
        self.parts = [(name, value)]
        self.part_type = part_type
        self.cluster = cluster
        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_next_record()
        self._atlas_entity_iterator = self._create_next_atlas_entity()

    def __repr__(self) -> str:
        return f"Watermark(create_time={str(self.create_time)!r}, database={self.database!r}, " \
               f"schema={self.schema!r}, table={self.table!r}, parts={self.parts!r}, " \
               f"cluster={self.cluster!r}, part_type={self.part_type!r})"

    def create_next_node(self) -> Union[GraphNode, None]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def get_watermark_model_key(self) -> str:
        return Watermark.KEY_FORMAT.format(database=self.database,
                                           cluster=self.cluster,
                                           schema=self.schema,
                                           table=self.table,
                                           part_type=self.part_type)

    def get_metadata_model_key(self) -> str:
        return f'{self.database}://{self.cluster}.{self.schema}/{self.table}'

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create watermark nodes
        :return:
        """
        for part in self.parts:
            part_node = GraphNode(
                key=self.get_watermark_model_key(),
                label=Watermark.LABEL,
                attributes={
                    'partition_key': part[0],
                    'partition_value': part[1],
                    'create_time': self.create_time
                }
            )
            yield part_node

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        """
        Create relation map between watermark record with original table
        :return:
        """
        relation = GraphRelationship(
            start_key=self.get_watermark_model_key(),
            start_label=Watermark.LABEL,
            end_key=self.get_metadata_model_key(),
            end_label='Table',
            type=Watermark.WATERMARK_TABLE_RELATION_TYPE,
            reverse_type=Watermark.TABLE_WATERMARK_RELATION_TYPE,
            attributes={}
        )
        yield relation

    def _create_next_record(self) -> Iterator[RDSModel]:
        """
        Create watermark records
        """
        for part in self.parts:
            part_record = RDSTableWatermark(
                rk=self.get_watermark_model_key(),
                partition_key=part[0],
                partition_value=part[1],
                create_time=self.create_time,
                table_rk=self.get_metadata_model_key()
            )
            yield part_record

    def _create_atlas_partition_entity(self, spec: Tuple[str, str]) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self.get_watermark_model_key()),
            ('name', spec[1]),
            ('displayName', spec[1]),
            ('key', spec[0]),
            ('create_time', self.create_time)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        relationship_list = []  # type: ignore

        add_entity_relationship(
            relationship_list,
            'table',
            AtlasTableTypes.table,
            self.get_metadata_model_key()
        )

        entity = AtlasEntity(
            typeName=AtlasTableTypes.watermark,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=get_entity_relationships(relationship_list)
        )

        return entity

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        pass

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        for part in self.parts:
            yield self._create_atlas_partition_entity(part)

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Iterable


class BaseFailureHandler(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def can_skip_failure(self,
                         exception: Exception,
                         ) -> bool:
        pass


class HttpFailureSkipOnStatus(BaseFailureHandler):

    def __init__(self,
                 status_codes_to_skip: Iterable[int],
                 ) -> None:
        self._status_codes_to_skip = {v for v in status_codes_to_skip}

    def can_skip_failure(self,
                         exception: Exception,
                         ) -> bool:
        try:
            status_code: int = getattr(getattr(exception, 'response'), 'status_code')
            return status_code in self._status_codes_to_skip
        except AttributeError:
            pass

        return False



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import (
    Any, Dict, Iterator, List, Optional, Union,
)

import boto3
import jsonref
from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata

LOGGER = logging.getLogger(__name__)


class EventBridgeExtractor(Extractor):
    """
    Extracts the latest version of all schemas from a given AWS EventBridge schema registry
    """

    REGION_NAME_KEY = "region_name"
    REGISTRY_NAME_KEY = "registry_name"
    DEFAULT_CONFIG = ConfigFactory.from_dict(
        {REGION_NAME_KEY: "us-east-1", REGISTRY_NAME_KEY: "aws.events"}
    )

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(EventBridgeExtractor.DEFAULT_CONFIG)

        boto3.setup_default_session(
            region_name=conf.get(EventBridgeExtractor.REGION_NAME_KEY)
        )
        self._schemas = boto3.client("schemas")

        self._registry_name = conf.get(EventBridgeExtractor.REGISTRY_NAME_KEY)

        self._extract_iter: Union[None, Iterator] = None

    def extract(self) -> Union[TableMetadata, None]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter(self._registry_name)
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self) -> str:
        return "extractor.eventbridge"

    def _get_extract_iter(self, registry_name: str) -> Iterator[TableMetadata]:
        """
        It gets all the schemas and yields TableMetadata
        :return:
        """
        for schema_desc in self._get_raw_extract_iter(registry_name):
            if "Content" not in schema_desc:
                LOGGER.warning(
                    f"skipped malformatted schema: {jsonref.dumps(schema_desc)}"
                )
                continue

            content = jsonref.loads(schema_desc["Content"])

            if content.get("openapi", "") == "3.0.0":  # NOTE: OpenAPI 3.0
                title = content.get("info", {}).get("title", "")
                for schema_name, schema in (
                    content.get("components", {}).get("schemas", {}).items()
                ):
                    table = EventBridgeExtractor._build_table(
                        schema,
                        schema_name,
                        registry_name,
                        title,
                        content.get("description", None),
                    )

                    if table is None:
                        continue

                    yield table
            elif (
                content.get("$schema", "") == "http://json-schema.org/draft-04/schema#"
            ):  # NOTE: JSON Schema Draft 4
                title = content.get("title", "")

                for schema_name, schema in content.get("definitions", {}).items():
                    table = EventBridgeExtractor._build_table(
                        schema,
                        schema_name,
                        registry_name,
                        title,
                        schema.get("description", None),
                    )

                    if table is None:
                        continue

                    yield table

                table = EventBridgeExtractor._build_table(
                    content,
                    "Root",
                    registry_name,
                    title,
                    content.get("description", None),
                )

                if table is None:
                    continue

                yield table

            else:
                LOGGER.warning(
                    f"skipped unsupported schema format: {jsonref.dumps(schema_desc)}"
                )
                continue

    def _get_raw_extract_iter(self, registry_name: str) -> Iterator[Dict[str, Any]]:
        """
        Provides iterator of results row from schemas client
        :return:
        """
        schemas_descs = self._search_schemas(registry_name)
        return iter(schemas_descs)

    def _search_schemas(self, registry_name: str) -> List[Dict[str, Any]]:
        """
        Get all schemas descriptions.
        """
        schemas_names = []
        paginator = self._schemas.get_paginator("list_schemas")
        for result in paginator.paginate(RegistryName=registry_name):
            for schema in result["Schemas"]:
                schemas_names.append(schema["SchemaName"])

        schemas_descs = []
        for schema_name in schemas_names:
            schema_versions = []
            paginator = self._schemas.get_paginator("list_schema_versions")
            for result in paginator.paginate(
                RegistryName=registry_name, SchemaName=schema_name
            ):
                schema_versions += result["SchemaVersions"]
            latest_schema_version = EventBridgeExtractor._get_latest_schema_version(
                schema_versions
            )

            schema_desc = self._schemas.describe_schema(
                RegistryName=registry_name,
                SchemaName=schema_name,
                SchemaVersion=latest_schema_version,
            )

            schemas_descs.append(schema_desc)

        return schemas_descs

    @staticmethod
    def _build_table(
        schema: Dict[str, Any],
        schema_name: str,
        registry_name: str,
        title: str,
        description: str,
    ) -> Optional[TableMetadata]:
        columns = []
        for i, (column_name, properties) in enumerate(
            schema.get("properties", {}).items()
        ):
            columns.append(
                ColumnMetadata(
                    column_name,
                    properties.get("description", None),
                    EventBridgeExtractor._get_property_type(properties),
                    i,
                )
            )

        if len(columns) == 0:
            LOGGER.warning(
                f"skipped schema with primitive type: "
                f"{schema_name}: {jsonref.dumps(schema)}"
            )
            return None

        return TableMetadata(
            "eventbridge", registry_name, title, schema_name, description, columns,
        )

    @staticmethod
    def _get_latest_schema_version(schema_versions: List[Dict[str, Any]]) -> str:
        versions = []
        for info in schema_versions:
            version = int(info["SchemaVersion"])
            versions.append(version)
        return str(max(versions))

    @staticmethod
    def _get_property_type(schema: Dict) -> str:
        if "type" not in schema:
            return "object"

        if schema["type"] == "object":
            properties = [
                f"{name}:{EventBridgeExtractor._get_property_type(_schema)}"
                for name, _schema in schema.get("properties", {}).items()
            ]
            if len(properties) > 0:
                return "struct<" + ",".join(properties) + ">"
            return "struct<object>"
        elif schema["type"] == "array":
            items = EventBridgeExtractor._get_property_type(schema.get("items", {}))
            return "array<" + items + ">"
        else:
            if "format" in schema:
                return f"{schema['type']}[{schema['format']}]"
            return schema["type"]



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any, Optional

from pyhocon import ConfigTree

from databuilder.loader.base_loader import Loader

LOGGER = logging.getLogger(__name__)

CALLBACK_FUNCTION = 'callback_function'


def log_call_back(record: Optional[Any]) -> None:
    """
    A Sample callback function. Implement any function follows this function's signature to fit your needs.
    :param record:
    :return:
    """
    LOGGER.info('record: %s', record)


class GenericLoader(Loader):
    """
    Loader class to call back a function provided by user
    """

    def init(self, conf: ConfigTree) -> None:
        """
        Initialize file handlers from conf
        :param conf:
        """
        self.conf = conf
        self._callback_func = self.conf.get(CALLBACK_FUNCTION, log_call_back)

    def load(self, record: Optional[Any]) -> None:
        """
        Write record to function
        :param record:
        :return:
        """
        if not record:
            return

        self._callback_func(record)

    def close(self) -> None:
        pass

    def get_scope(self) -> str:
        return "loader.generic"



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from os import listdir
from os.path import isfile, join
from typing import (
    Iterator, List, Set,
)

import pandas
from jinja2 import Template
from neo4j import Neo4jDriver, Transaction
from neo4j.exceptions import Neo4jError
from pyhocon import ConfigTree

from databuilder.models.graph_serializable import NODE_LABEL
from databuilder.publisher.publisher_config_constants import PublisherConfigs

LOGGER = logging.getLogger(__name__)

NEO4J_EQUIVALENT_SCHEMA_RULE_ALREADY_EXISTS_ERROR_CODE = 'Neo.ClientError.Schema.EquivalentSchemaRuleAlreadyExists'
NEO4J_INDEX_ALREADY_EXISTS_ERROR_CODE = 'Neo.ClientError.Schema.IndexWithNameAlreadyExists'


def chunkify_list(records: List[dict], chunk_size: int) -> Iterator[List[dict]]:
    """
    Generator to evenly split the input list into chunks
    """
    for index in range(0, len(records), chunk_size):
        yield records[index:index + chunk_size]


def create_neo4j_node_key_constraint(node_file: str,
                                     current_labels: Set,
                                     driver: Neo4jDriver,
                                     db_name: str) -> Set:
    """
    Go over the node file and try creating unique indices.
    For any label seen first time for this publisher it will try to create unique index.
    Neo4j ignores a second creation in 3.x, but raises an error in 4.x.
    """
    LOGGER.info('Creating indices using Node file: %s. (Existing indices will be ignored)', node_file)

    labels = set(current_labels)
    with open(node_file, 'r', encoding='utf8') as node_csv:
        for node_record in pandas.read_csv(node_csv,
                                           na_filter=False).to_dict(orient='records'):
            label = node_record[NODE_LABEL]
            if label not in labels:
                with driver.session(database=db_name) as session:
                    try:
                        create_stmt = Template("""
                            CREATE CONSTRAINT ON (node:{{ LABEL }}) ASSERT node.key IS UNIQUE
                        """).render(LABEL=label)

                        LOGGER.info(f'Trying to create index for label {label} if not exist: {create_stmt}')

                        session.write_transaction(execute_neo4j_statement, create_stmt)
                    except Neo4jError as e:
                        if e.code != NEO4J_EQUIVALENT_SCHEMA_RULE_ALREADY_EXISTS_ERROR_CODE\
                                and e.code != NEO4J_INDEX_ALREADY_EXISTS_ERROR_CODE:
                            raise
                        # Else, swallow the exception, to make this function idempotent.
                labels.add(label)

    LOGGER.info('Indices have been created.')
    return labels


def create_props_param(record_dict: dict, additional_publisher_metadata_fields: dict) -> dict:
    """
    Create a dict of all the params for a given record
    """
    params = {}

    for k, v in {**record_dict, **additional_publisher_metadata_fields}.items():
        params[strip_unquoted_suffix(k)] = v

    return params


def execute_neo4j_statement(tx: Transaction,
                            stmt: str,
                            params: dict = None) -> None:
    """
    Executes statement against Neo4j. If execution fails, it rollsback and raises exception.
    """
    LOGGER.debug('Executing statement: %s with params %s', stmt, params)

    tx.run(stmt, parameters=params)


def get_props_body_keys(record_keys: list,
                        exclude_keys: Set,
                        additional_publisher_metadata_fields: dict) -> Set:
    """
    Returns the set of keys to be used in the props body of the merge statements
    :param record_keys:
    :param exclude_keys: set of excluded columns that do not need to be in properties (e.g: KEY, LABEL ...)
    :param additional_publisher_metadata_fields:
    """
    props_body_keys = set(record_keys) - exclude_keys
    formatted_keys = map(strip_unquoted_suffix, props_body_keys)
    return set(formatted_keys).union(additional_publisher_metadata_fields.keys())


def list_files(conf: ConfigTree, path_key: str) -> List[str]:
    """
    List files from directory
    :param conf:
    :param path_key:
    :return: List of file paths
    """
    if path_key not in conf:
        return []

    path = conf.get_string(path_key)
    return [join(path, f) for f in listdir(path) if isfile(join(path, f))]


def strip_unquoted_suffix(key: str) -> str:
    return key[:-len(PublisherConfigs.UNQUOTED_SUFFIX)] if key.endswith(PublisherConfigs.UNQUOTED_SUFFIX) else key



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging

from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_executions_extractor import (
    ModeDashboardExecutionsExtractor,
)
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_utils import ModeDashboardUtils
from databuilder.extractor.restapi.rest_api_extractor import STATIC_RECORD_DICT
from databuilder.models.dashboard.dashboard_execution import DashboardExecution
from databuilder.rest_api.mode_analytics.mode_paginated_rest_api_query import ModePaginatedRestApiQuery

LOGGER = logging.getLogger(__name__)


class ModeDashboardLastSuccessfulExecutionExtractor(ModeDashboardExecutionsExtractor):
    """
    A Extractor that extracts Mode dashboard's last successful run (execution) timestamp.

    """

    def __init__(self) -> None:
        super(ModeDashboardLastSuccessfulExecutionExtractor, self).__init__()

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(
            ConfigFactory.from_dict({
                STATIC_RECORD_DICT: {'product': 'mode',
                                     'execution_state': 'succeeded',
                                     'execution_id': DashboardExecution.LAST_SUCCESSFUL_EXECUTION_ID}
            })
        )
        super(ModeDashboardLastSuccessfulExecutionExtractor, self).init(conf)

    def get_scope(self) -> str:
        return 'extractor.mode_dashboard_last_successful_execution'

    def _build_restapi_query(self) -> ModePaginatedRestApiQuery:
        """
        Build REST API Query to get Mode Dashboard last successful execution metadata.
        :return: A RestApiQuery that provides Mode Dashboard last successful execution (run)
        """

        seed_query = ModeDashboardUtils.get_seed_query(conf=self._conf)
        params = ModeDashboardUtils.get_auth_params(conf=self._conf, discover_auth=True)

        # Reports
        # https://mode.com/developer/discovery-api/analytics/reports/
        url = 'https://app.mode.com/batch/{organization}/reports'
        json_path = 'reports[*].[token, space_token, last_successfully_run_at]'
        field_names = ['dashboard_id', 'dashboard_group_id', 'execution_timestamp']
        max_record_size = 1000
        pagination_json_path = 'reports[*]'
        last_successful_run_query = ModePaginatedRestApiQuery(query_to_join=seed_query, url=url, params=params,
                                                              json_path=json_path, field_names=field_names,
                                                              skip_no_result=True, max_record_size=max_record_size,
                                                              pagination_json_path=pagination_json_path)

        return last_successful_run_query



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any, Dict, Iterator, List, Optional, Set,
)

from databuilder.models.description_metadata import DescriptionMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_metadata import (
    TableMetadata, TagMetadata, _format_as_list,
)


class FeatureMetadata(GraphSerializable):
    """
    Base feature metadata.

    It implements GraphSerializable (TODO: implement TableSerializable)
    so that it can be serialized to produce Feature, Feature_Group, Tag,
    Database, Description and relations between those.
    """

    NODE_LABEL = 'Feature'
    KEY_FORMAT = '{feature_group}/{name}/{version}'

    NAME_ATTR = 'name'
    VERSION_ATTR = 'version'
    STATUS_ATTR = 'status'
    ENTITY_ATTR = 'entity'
    DATA_TYPE_ATTR = 'data_type'
    CREATED_TIMESTAMP_ATTR = 'created_timestamp'
    LAST_UPDATED_TIMESTAMP_ATTR = 'last_updated_timestamp'

    GROUP_NODE_LABEL = 'Feature_Group'
    GROUP_KEY_FORMAT = '{feature_group}'
    GROUP_FEATURE_RELATION_TYPE = 'GROUPS'
    FEATURE_GROUP_RELATION_TYPE = 'GROUPED_BY'

    FEATURE_DATABASE_RELATION_TYPE = 'FEATURE_AVAILABLE_IN'
    DATABASE_FEATURE_RELATION_TYPE = 'AVAILABLE_FEATURE'

    processed_feature_group_keys: Set[str] = set()
    processed_database_keys: Set[str] = set()

    def __init__(self,
                 feature_group: str,
                 name: str,
                 version: str,
                 status: Optional[str] = None,
                 entity: Optional[str] = None,
                 data_type: Optional[str] = None,
                 availability: List[str] = None,  # list of databases
                 description: Optional[str] = None,
                 tags: List[str] = None,
                 created_timestamp: Optional[int] = None,
                 last_updated_timestamp: Optional[int] = None,
                 **kwargs: Any
                 ) -> None:

        self.feature_group = feature_group
        self.name = name
        self.version = version
        self.status = status
        # what business entity the feature is about, e.g. 'Buyer', 'Ride', 'Listing', etc.
        self.entity = entity
        self.data_type = data_type
        self.availability = _format_as_list(availability)
        self.description = DescriptionMetadata.create_description_metadata(text=description)
        self.tags = _format_as_list(tags)
        self.created_timestamp = created_timestamp
        self.last_updated_timestamp = last_updated_timestamp

        self._node_iterator = self._create_next_node()
        self._relation_iterator = self._create_next_relation()

    def __repr__(self) -> str:
        return f'FeatureMetadata(' \
               f'{self.feature_group!r}, {self.name!r}, {self.version!r}, {self.status!r}, ' \
               f'{self.entity!r}, {self.data_type!r}, {self.availability!r}, {self.description!r}, ' \
               f'{self.tags!r}, {self.created_timestamp!r}, {self.last_updated_timestamp!r})'

    def _get_feature_key(self) -> str:
        return FeatureMetadata.KEY_FORMAT.format(feature_group=self.feature_group,
                                                 name=self.name,
                                                 version=self.version)

    def _get_feature_group_key(self) -> str:
        return FeatureMetadata.GROUP_KEY_FORMAT.format(feature_group=self.feature_group)

    def create_next_node(self) -> Optional[GraphNode]:
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _get_feature_node_attributes(self) -> Dict[str, Any]:
        feature_node_attrs: Dict[str, Any] = {
            FeatureMetadata.NAME_ATTR: self.name,
        }
        if self.version:
            feature_node_attrs[FeatureMetadata.VERSION_ATTR] = self.version

        if self.status:
            feature_node_attrs[FeatureMetadata.STATUS_ATTR] = self.status

        if self.entity:
            feature_node_attrs[FeatureMetadata.ENTITY_ATTR] = self.entity

        if self.data_type:
            feature_node_attrs[FeatureMetadata.DATA_TYPE_ATTR] = self.data_type

        if self.created_timestamp:
            feature_node_attrs[FeatureMetadata.CREATED_TIMESTAMP_ATTR] = self.created_timestamp

        if self.last_updated_timestamp:
            feature_node_attrs[FeatureMetadata.LAST_UPDATED_TIMESTAMP_ATTR] = self.last_updated_timestamp

        return feature_node_attrs

    def _create_next_node(self) -> Iterator[GraphNode]:
        yield GraphNode(
            key=self._get_feature_key(),
            label=FeatureMetadata.NODE_LABEL,
            attributes=self._get_feature_node_attributes()
        )

        if self.feature_group:
            fg = GraphNode(
                key=self._get_feature_group_key(),
                label=FeatureMetadata.GROUP_NODE_LABEL,
                attributes={FeatureMetadata.NAME_ATTR: self.feature_group},
            )
            if fg.key not in FeatureMetadata.processed_feature_group_keys:
                FeatureMetadata.processed_feature_group_keys.add(fg.key)
                yield fg

        if self.description:
            yield self.description.get_node(
                node_key=self.description.get_description_default_key(  # type: ignore
                    start_key=self._get_feature_key()),
            )

        for database in self.availability:
            db = GraphNode(
                key=TableMetadata.DATABASE_KEY_FORMAT.format(db=database),
                label=TableMetadata.DATABASE_NODE_LABEL,
                attributes={
                    FeatureMetadata.NAME_ATTR: database
                }
            )
            if db.key not in FeatureMetadata.processed_database_keys:
                FeatureMetadata.processed_database_keys.add(db.key)
                yield db

        for tag_value in self.tags:
            yield TagMetadata(name=tag_value).get_node()

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_next_relation(self) -> Iterator[GraphRelationship]:
        # Feature <> Feature group
        if self.feature_group:
            yield GraphRelationship(
                start_label=FeatureMetadata.NODE_LABEL,
                end_label=FeatureMetadata.GROUP_NODE_LABEL,
                start_key=self._get_feature_key(),
                end_key=self._get_feature_group_key(),
                type=FeatureMetadata.FEATURE_GROUP_RELATION_TYPE,
                reverse_type=FeatureMetadata.GROUP_FEATURE_RELATION_TYPE,
                attributes={}
            )

        # Feature <> Description
        if self.description:
            yield GraphRelationship(
                start_label=FeatureMetadata.NODE_LABEL,
                end_label=DescriptionMetadata.DESCRIPTION_NODE_LABEL,
                start_key=self._get_feature_key(),
                end_key=self.description.get_description_default_key(start_key=self._get_feature_key()),
                type=DescriptionMetadata.DESCRIPTION_RELATION_TYPE,
                reverse_type=DescriptionMetadata.INVERSE_DESCRIPTION_RELATION_TYPE,
                attributes={}
            )

        # Feature <> Database
        for database in self.availability:
            yield GraphRelationship(
                start_label=FeatureMetadata.NODE_LABEL,
                end_label=TableMetadata.DATABASE_NODE_LABEL,
                start_key=self._get_feature_key(),
                end_key=TableMetadata.DATABASE_KEY_FORMAT.format(db=database),
                type=FeatureMetadata.FEATURE_DATABASE_RELATION_TYPE,
                reverse_type=FeatureMetadata.DATABASE_FEATURE_RELATION_TYPE,
                attributes={}
            )

        # Feature <> Tag
        for tag in self.tags:
            yield GraphRelationship(
                start_label=FeatureMetadata.NODE_LABEL,
                end_label=TagMetadata.TAG_NODE_LABEL,
                start_key=self._get_feature_key(),
                end_key=TagMetadata.get_tag_key(tag),
                type=TagMetadata.ENTITY_TAG_RELATION_TYPE,
                reverse_type=TagMetadata.TAG_ENTITY_RELATION_TYPE,
                attributes={}
            )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import os

from setuptools import find_packages, setup

__version__ = '7.3.0'

requirements_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                 'requirements.txt')
with open(requirements_path, 'r') as requirements_file:
    requirements = requirements_file.readlines()

requirements_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                 'requirements-dev.txt')
with open(requirements_path, 'r') as requirements_file:
    requirements_dev = requirements_file.readlines()

kafka = ['confluent-kafka==1.0.0']

cassandra = ['cassandra-driver==3.20.1']

glue = ['boto3==1.17.23']

snowflake = [
    'snowflake-connector-python',
    'snowflake-sqlalchemy'
]

athena = ['PyAthena[SQLAlchemy]>=1.0.0, <2.0.0']

# Python API client for google
# License: Apache Software License
# Upstream url: https://github.com/googleapis/google-api-python-client
bigquery = [
    'google-api-python-client>=1.6.0, <2.0.0dev',
    'google-auth-httplib2>=0.0.1',
    'google-auth>=1.16.0, <3.0.0dev'
]

jsonpath = ['jsonpath_rw==1.4.0']

db2 = [
    'ibm_db>=3.0.1',
    'ibm-db-sa-py3>=0.3.1-1'
]

dremio = [
    'pyodbc==4.0.30'
]

druid = [
    'pydruid'
]

spark = [
    'pyspark == 3.0.1'
]

neptune = [
    'amundsen-gremlin>=0.0.9',
    'Flask==1.0.2',
    'gremlinpython==3.4.3',
    'requests-aws4auth==1.1.0',
    'typing-extensions==4.0.0',
    'overrides==2.5',
    'boto3==1.17.23'
]

feast = [
    'feast==0.17.0',
    'fastapi!=0.76.*'
]

atlas = [
    'pyatlasclient>=1.1.2',
    'apache-atlas>=0.0.11'
]

oracle = [
    'cx_Oracle==8.2.1'
]

rds = [
    'sqlalchemy>=1.3.6,<1.4',
    'mysqlclient>=1.3.6,<3'
]

salesforce = [
    'simple-salesforce>=1.11.2'
]

teradata = [
    'teradatasqlalchemy==17.0.0.0'
]

schema_registry = [
    'python-schema-registry-client==2.4.0'
]

all_deps = requirements + requirements_dev + kafka + cassandra + glue + snowflake + athena + \
    bigquery + jsonpath + db2 + dremio + druid + spark + feast + neptune + rds \
    + atlas + salesforce + oracle + teradata + schema_registry

setup(
    name='amundsen-databuilder',
    version=__version__,
    description='Amundsen Data builder',
    url='https://www.github.com/amundsen-io/amundsen/tree/main/databuilder',
    maintainer='Amundsen TSC',
    maintainer_email='amundsen-tsc@lists.lfai.foundation',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    dependency_links=[],
    install_requires=requirements,
    python_requires='>=3.7',
    extras_require={
        'all': all_deps,
        'dev': requirements_dev,
        'kafka': kafka,  # To use with Kafka source extractor
        'cassandra': cassandra,
        'glue': glue,
        'snowflake': snowflake,
        'athena': athena,
        'bigquery': bigquery,
        'jsonpath': jsonpath,
        'db2': db2,
        'dremio': dremio,
        'druid': druid,
        'neptune': neptune,
        'delta': spark,
        'feast': feast,
        'atlas': atlas,
        'rds': rds,
        'salesforce': salesforce,
        'oracle': oracle,
        'teradata': teradata,
        'schema_registry': schema_registry,
    },
    classifiers=[
        'Programming Language :: Python :: 3.7',
    ],
)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from typing import (
    Dict, Iterator, Optional, Tuple, Union,
)

from databuilder.extractor.es_base_extractor import ElasticsearchBaseExtractor
from databuilder.models.watermark import Watermark


class ElasticsearchWatermarkExtractor(ElasticsearchBaseExtractor):
    """
    Extractor to extract index watermarks from Elasticsearch
    """

    def get_scope(self) -> str:
        return 'extractor.es_watermark'

    # Internally, Elasticsearch stores dates as numbers representing milliseconds since the epoch,
    # so the agg result is expected to be floats.
    # See https://www.elastic.co/guide/en/elasticsearch/reference/current/date.html#date
    def _get_index_watermark_bounds(self, index_name: str) -> Optional[Tuple[float, float]]:
        try:
            search_result = self.es.search(
                index=index_name,
                size=0,
                aggs={
                    'min_watermark': {'min': {'field': self._time_field}},
                    'max_watermark': {'max': {'field': self._time_field}}
                }
            )
            watermark_min = search_result.get('aggregations').get('min_watermark').get('value')
            watermark_max = search_result.get('aggregations').get('max_watermark').get('value')
            if watermark_min is not None and watermark_max is not None:
                return float(watermark_min), float(watermark_max)
        except Exception:
            pass

        return None

    def _get_extract_iter(self) -> Iterator[Union[Watermark, None]]:
        # Get all the indices
        indices: Dict = self._get_indexes()

        # Iterate over indices
        for index_name, index_metadata in indices.items():
            creation_date: Optional[float] = self._get_index_creation_date(index_metadata)
            watermark_bounds: Optional[Tuple[float, float]] = self._get_index_watermark_bounds(index_name=index_name)
            watermark_min: Optional[float] = None if watermark_bounds is None else watermark_bounds[0]
            watermark_max: Optional[float] = None if watermark_bounds is None else watermark_bounds[1]

            if creation_date is None or watermark_min is None or watermark_max is None:
                continue

            creation_date_str: str = datetime.fromtimestamp(creation_date / 1000).strftime('%Y-%m-%d %H:%M:%S')
            watermark_min_str: str = datetime.fromtimestamp(watermark_min / 1000).strftime('%Y-%m-%d')
            watermark_max_str: str = datetime.fromtimestamp(watermark_max / 1000).strftime('%Y-%m-%d')

            yield Watermark(
                database=self.database,
                cluster=self.cluster,
                schema=self.schema,
                table_name=index_name,
                create_time=creation_date_str,
                part_name=f'{self._time_field}={watermark_min_str}',
                part_type='low_watermark'
            )

            yield Watermark(
                database=self.database,
                cluster=self.cluster,
                schema=self.schema,
                table_name=index_name,
                create_time=creation_date_str,
                part_name=f'{self._time_field}={watermark_max_str}',
                part_type='high_watermark'
            )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import copy
from typing import (
    Any, Iterator, Optional, Union,
)

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasCommonTypes
from amundsen_rds.models import RDSModel
from amundsen_rds.models.user import User as RDSUser

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers.atlas_serializer import get_entity_attrs
from databuilder.utils.atlas import AtlasSerializedEntityOperation


class User(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    User model. This model doesn't define any relationship.
    """
    USER_NODE_LABEL = 'User'
    USER_NODE_KEY_FORMAT = '{email}'
    USER_NODE_EMAIL = 'email'
    USER_NODE_FIRST_NAME = 'first_name'
    USER_NODE_LAST_NAME = 'last_name'
    USER_NODE_FULL_NAME = 'full_name'
    USER_NODE_GITHUB_NAME = 'github_username'
    USER_NODE_TEAM = 'team_name'
    USER_NODE_EMPLOYEE_TYPE = 'employee_type'
    USER_NODE_MANAGER_EMAIL = 'manager_email'
    USER_NODE_SLACK_ID = 'slack_id'
    USER_NODE_IS_ACTIVE = 'is_active'  # bool value needs to be unquoted when publish to neo4j
    USER_NODE_PROFILE_URL = 'profile_url'
    USER_NODE_UPDATED_AT = 'updated_at'
    USER_NODE_ROLE_NAME = 'role_name'

    USER_MANAGER_RELATION_TYPE = 'MANAGE_BY'
    MANAGER_USER_RELATION_TYPE = 'MANAGE'

    def __init__(self,
                 email: str,
                 first_name: str = '',
                 last_name: str = '',
                 full_name: str = '',
                 github_username: str = '',
                 team_name: str = '',
                 employee_type: str = '',
                 manager_email: str = '',
                 slack_id: str = '',
                 is_active: bool = True,
                 profile_url: str = '',
                 updated_at: int = 0,
                 role_name: str = '',
                 do_not_update_empty_attribute: bool = False,
                 **kwargs: Any
                 ) -> None:
        """
        This class models user node for Amundsen people.

        :param first_name:
        :param last_name:
        :param full_name:
        :param email:
        :param github_username:
        :param team_name:
        :param employee_type:
        :param manager_email:
        :param is_active:
        :param profile_url:
        :param updated_at: everytime we update the node, we will push the timestamp.
                           then we will have a cron job to update the ex-employee nodes based on
                           the case if this timestamp hasn't been updated for two weeks.
        :param role_name: the role_name of the user (e.g swe)
        :param do_not_update_empty_attribute: If False, all empty or not defined params will be overwritten with
        empty string.
        :param kwargs: Any K/V attributes we want to update the
        """
        self.first_name = first_name
        self.last_name = last_name
        self.full_name = full_name

        self.email = email
        self.github_username = github_username
        # todo: team will be a separate node once Amundsen People supports team
        self.team_name = team_name
        self.manager_email = manager_email
        self.employee_type = employee_type
        # this attr not available in team service, either update team service, update with FE
        self.slack_id = slack_id
        self.is_active = is_active
        self.profile_url = profile_url
        self.updated_at = updated_at
        self.role_name = role_name
        self.do_not_update_empty_attribute = do_not_update_empty_attribute
        self.attrs = None
        if kwargs:
            self.attrs = copy.deepcopy(kwargs)

        self._node_iter = self._create_node_iterator()
        self._rel_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()

    def create_next_node(self) -> Optional[GraphNode]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        """
        :return:
        """
        try:
            return next(self._rel_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    @classmethod
    def get_user_model_key(cls,
                           email: str = None
                           ) -> str:
        if not email:
            return ''
        return User.USER_NODE_KEY_FORMAT.format(email=email)

    def get_user_node(self) -> GraphNode:
        node_attributes = {
            User.USER_NODE_EMAIL: self.email,
            User.USER_NODE_IS_ACTIVE: self.is_active,
            User.USER_NODE_PROFILE_URL: self.profile_url or '',
            User.USER_NODE_FIRST_NAME: self.first_name or '',
            User.USER_NODE_LAST_NAME: self.last_name or '',
            User.USER_NODE_FULL_NAME: self.full_name or '',
            User.USER_NODE_GITHUB_NAME: self.github_username or '',
            User.USER_NODE_TEAM: self.team_name or '',
            User.USER_NODE_EMPLOYEE_TYPE: self.employee_type or '',
            User.USER_NODE_SLACK_ID: self.slack_id or '',
            User.USER_NODE_ROLE_NAME: self.role_name or ''
        }

        if self.updated_at:
            node_attributes[User.USER_NODE_UPDATED_AT] = self.updated_at
        elif not self.do_not_update_empty_attribute:
            node_attributes[User.USER_NODE_UPDATED_AT] = 0

        if self.attrs:
            for k, v in self.attrs.items():
                if k not in node_attributes:
                    node_attributes[k] = v

        if self.do_not_update_empty_attribute:
            for k, v in list(node_attributes.items()):
                if not v:
                    del node_attributes[k]

        node = GraphNode(
            key=User.get_user_model_key(email=self.email),
            label=User.USER_NODE_LABEL,
            attributes=node_attributes
        )

        return node

    def get_user_record(self) -> RDSModel:
        record_attr_map = {
            RDSUser.email: self.email,
            RDSUser.is_active: self.is_active,
            RDSUser.profile_url: self.profile_url or '',
            RDSUser.first_name: self.first_name or '',
            RDSUser.last_name: self.last_name or '',
            RDSUser.full_name: self.full_name or '',
            RDSUser.github_username: self.github_username or '',
            RDSUser.team_name: self.team_name or '',
            RDSUser.employee_type: self.employee_type or '',
            RDSUser.slack_id: self.slack_id or '',
            RDSUser.role_name: self.role_name or '',
            RDSUser.updated_at: self.updated_at or 0
        }

        record = RDSUser(rk=User.get_user_model_key(email=self.email))
        # set value for attributes of user record if the value is not empty
        # or the flag allows to update empty values
        for attr, value in record_attr_map.items():
            if value or not self.do_not_update_empty_attribute:
                record.__setattr__(attr.key, value)

        if self.manager_email:
            record.manager_rk = self.get_user_model_key(email=self.manager_email)

        return record

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create an user node
        :return:
        """
        user_node = self.get_user_node()
        yield user_node

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        if self.manager_email:
            # only create the relation if the manager exists
            relationship = GraphRelationship(
                start_key=User.get_user_model_key(email=self.email),
                start_label=User.USER_NODE_LABEL,
                end_label=User.USER_NODE_LABEL,
                end_key=self.get_user_model_key(email=self.manager_email),
                type=User.USER_MANAGER_RELATION_TYPE,
                reverse_type=User.MANAGER_USER_RELATION_TYPE,
                attributes={}
            )
            yield relationship

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        user_record = self.get_user_record()
        yield user_record

    def _create_atlas_user_entity(self) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, User.get_user_model_key(email=self.email)),
            ('email', self.email),
            ('first_name', self.first_name),
            ('last_name', self.last_name),
            ('full_name', self.full_name),
            ('github_username', self.github_username),
            ('team_name', self.team_name),
            ('employee_type', self.employee_type),
            ('manager_email', self.manager_email),
            ('slack_id', self.slack_id),
            ('is_active', self.is_active),
            ('profile_url', self.profile_url),
            ('updated_at', self.updated_at),
            ('role_name', self.role_name),
            ('displayName', self.email)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        entity = AtlasEntity(
            typeName=AtlasCommonTypes.user,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=None
        )

        return entity

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        pass

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        yield self._create_atlas_user_entity()

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None

    def __repr__(self) -> str:
        return f'User({self.first_name!r}, {self.last_name!r}, {self.full_name!r}, {self.email!r}, ' \
               f'{self.github_username!r}, {self.team_name!r}, {self.slack_id!r}, {self.manager_email!r}, ' \
               f'{self.employee_type!r}, {self.is_active!r}, {self.profile_url!r}, {self.updated_at!r}, ' \
               f'{self.role_name!r})'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional

from databuilder.models.elasticsearch_document import ElasticsearchDocument


class FeatureESDocument(ElasticsearchDocument):
    """
    Schema for the Feature ES document
    """

    def __init__(self,
                 feature_group: str,
                 feature_name: str,
                 version: str,
                 key: str,
                 total_usage: int,
                 status: Optional[str] = None,
                 entity: Optional[str] = None,
                 description: Optional[str] = None,
                 availability: Optional[List[str]] = None,
                 badges: Optional[List[str]] = None,
                 tags: Optional[List[str]] = None,
                 last_updated_timestamp: Optional[int] = None,
                 ) -> None:
        self.feature_group = feature_group
        self.feature_name = feature_name
        self.version = version
        self.key = key
        self.total_usage = total_usage
        self.status = status
        self.entity = entity
        self.description = description
        self.availability = availability
        self.badges = badges
        self.tags = tags
        self.last_updated_timestamp = last_updated_timestamp



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0


from typing import (
    Any, Dict, Iterator, Union,
)

from databuilder.extractor.dashboard.apache_superset.apache_superset_extractor import (
    ApacheSupersetBaseExtractor, type_fields_mapping,
)
from databuilder.models.dashboard.dashboard_last_modified import DashboardLastModifiedTimestamp
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata


class ApacheSupersetMetadataExtractor(ApacheSupersetBaseExtractor):
    def last_modified_field_mappings(self) -> type_fields_mapping:
        result = [
            ('dashboard_id', 'result.id', lambda x: str(x), ''),
            ('dashboard_name', 'result.dashboard_title', None, ''),
            ('last_modified_timestamp', 'result.changed_on',
             lambda _date: self.parse_date(_date), 0)
        ]

        return result

    def dashboard_metadata_field_mappings(self) -> type_fields_mapping:
        result = [
            ('dashboard_id', 'result.id', lambda x: str(x), ''),
            ('dashboard_name', 'result.dashboard_title', None, ''),
            ('dashboard_url', 'result.url', lambda x: self.base_url + x, ''),
            ('created_timestamp', 'result.created_on', None, 0),  # currently not available in superset dashboard api
            ('tags', '', lambda x: [x] if x else [], []),  # not available
            ('description', 'result.description', None, '')  # currently not available in superset dashboard api
        ]

        return result

    def _get_extract_iter(self) -> Iterator[Union[DashboardMetadata, DashboardLastModifiedTimestamp, None]]:
        ids = self._get_resource_ids('dashboard')

        data = [self._get_dashboard_details(i) for i in ids]

        if self.extract_published_only:
            data = [d for d in data if self.get_nested_field(d, 'result.published')]

        for entry in data:
            dashboard_metadata = self.map_fields(entry, self.dashboard_metadata_field_mappings())
            dashboard_metadata.update(**self.common_params)

            yield DashboardMetadata(**dashboard_metadata)

            dashboard_last_modified = self.map_fields(entry, self.last_modified_field_mappings())
            dashboard_last_modified.update(**self.common_params)

            yield DashboardLastModifiedTimestamp(**dashboard_last_modified)

    def _get_dashboard_details(self, dashboard_id: str) -> Dict[str, Any]:
        url = self.build_full_url(f'api/v1/dashboard/{dashboard_id}')

        data = self.execute_query(url)

        return data
